"""
SimSiam network architecture for 3D brain volumes
Ref: [Simple Siamese Representation Learning by Chen et al.](https://arxiv.org/abs/2011.10566).
author: Dhritiman Das
"""

import tensorflow as tf
from tensorflow.keras import layers, regularizers

import nobrainer


def brainsiam(
    n_classes,
    input_shape,
    weight_decay=0.0005,
    projection_dim=2048,
    latent_dim=512,
    name="brainsiam",
    **kwargs
):
    """Instantiate Brain Siamese Network model."""

    """ Parameters
    ------------------
    input_shape: list or tuple of four ints, the shape of the input data. Should be
        scaled to [0,1]. Omit the batch dimension, and include the number of channels.
        Currently, only squares and cubes supported.
    projection_dim: int, the dimensions of the encoding of the input data.
    latent_dim: int, the dimensions of the latent space of the input data.
    n_classes: int, the number of classes in the input data.
    weight_decay: float, rate of decay for weights in the the l2 regularizer

    Returns
    -------
    2 Model objects: encoder, predictor.
    """

    print("projection dimension is: ", projection_dim)
    print("latent dimension: ", latent_dim)

    def encoder():
        resnet = nobrainer.models.highresnet(
            n_classes=n_classes, input_shape=input_shape
        )

        input = tf.keras.layers.Input(shape=input_shape)

        resnet_out = resnet(input)

        x = layers.GlobalAveragePooling3D(name="backbone_pool")(resnet_out)

        x = layers.Dense(
            projection_dim,
            use_bias=False,
            kernel_regularizer=regularizers.l2(weight_decay),
        )(x)
        x = layers.BatchNormalization()(x)
        x = layers.LeakyReLU()(x)
        x = layers.Dense(
            projection_dim,
            use_bias=False,
            kernel_regularizer=regularizers.l2(weight_decay),
        )(x)
        output = layers.BatchNormalization()(x)

        encoder_model = tf.keras.Model(input, output, name="encoder")
        return encoder_model

    def predictor():
        predictor_model = tf.keras.Sequential(
            [
                # Note the AutoEncoder-like structure.
                tf.keras.layers.InputLayer((projection_dim,)),
                tf.keras.layers.Dense(
                    latent_dim,
                    use_bias=False,
                    kernel_regularizer=regularizers.l2(weight_decay),
                ),
                tf.keras.layers.LeakyReLU(),
                tf.keras.layers.BatchNormalization(),
                tf.keras.layers.Dense(projection_dim),
            ],
            name="predictor",
        )

        return predictor_model

    encoder_model = encoder()
    predictor_model = predictor()

    encoder_model.summary()
    predictor_model.summary()

    return encoder_model, predictor_model



import gzip
import tempfile

from fsspec.implementations.local import LocalFileSystem
import nibabel as nib
import numpy as np
import pytest

from .utils import csv_of_volumes  # noqa: F401
from .. import io


def test_read_csv():
    with tempfile.NamedTemporaryFile() as f:
        f.write("foo,bar\nbaz,boo".encode())
        f.seek(0)
        assert [("foo", "bar"), ("baz", "boo")] == io.read_csv(
            f.name, skip_header=False
        )

    with tempfile.NamedTemporaryFile() as f:
        f.write("foo,bar\nbaz,boo".encode())
        f.seek(0)
        assert [("baz", "boo")] == io.read_csv(f.name, skip_header=True)

    with tempfile.NamedTemporaryFile() as f:
        f.write("foo,bar\nbaz,boo".encode())
        f.seek(0)
        assert [("baz", "boo")] == io.read_csv(f.name)

    with tempfile.NamedTemporaryFile() as f:
        f.write("foo|bar\nbaz|boo".encode())
        f.seek(0)
        assert [("baz", "boo")] == io.read_csv(f.name, delimiter="|")


def test_read_mapping():
    with tempfile.NamedTemporaryFile() as f:
        f.write("orig,new\n0,1\n20,10\n40,15".encode())
        f.seek(0)
        assert {0: 1, 20: 10, 40: 15} == io.read_mapping(f.name, skip_header=True)
        # Header is non-integer.
        with pytest.raises(ValueError):
            io.read_mapping(f.name, skip_header=False)

    with tempfile.NamedTemporaryFile() as f:
        f.write("orig,new\n0,1\n20,10\n40".encode())
        f.seek(0)
        # Last row only has one value.
        with pytest.raises(ValueError):
            io.read_mapping(f.name, skip_header=False)

    with tempfile.NamedTemporaryFile() as f:
        f.write("origFnew\n0F1\n20F10\n40F15".encode())
        f.seek(0)
        assert {0: 1, 20: 10, 40: 15} == io.read_mapping(
            f.name, skip_header=True, delimiter="F"
        )


def test_read_volume(tmp_path):
    data = np.random.rand(8, 8, 8).astype(np.float32)
    affine = np.eye(4)

    filename = str(tmp_path / "foo.nii.gz")
    nib.save(nib.Nifti1Image(data, affine), filename)
    data_loaded = io.read_volume(filename)
    assert np.array_equal(data, data_loaded)

    data_loaded = io.read_volume(filename, dtype=data.dtype)
    assert data.dtype == data_loaded.dtype

    data_loaded, affine_loaded = io.read_volume(filename, return_affine=True)
    assert np.array_equal(data, data_loaded)
    assert np.array_equal(affine, affine_loaded)

    data = np.random.rand(8, 8, 8).astype(np.float32)
    affine = np.array([[1.5, 0, 1.2, 0], [0.8, 0.8, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1]])
    filename = str(tmp_path / "foo_asr.nii.gz")
    nib.save(nib.Nifti1Image(data, affine), filename)
    data_loaded = io.read_volume(filename, to_ras=True)
    assert not np.array_equal(data, data_loaded)
    data_loaded = io.read_volume(filename, to_ras=False)
    assert np.array_equal(data, data_loaded)


def test_verify_features_nonscalar_labels(csv_of_volumes):  # noqa: F811
    files = io.read_csv(csv_of_volumes, skip_header=False)
    invalid = io.verify_features_labels(
        files, volume_shape=(8, 8, 8), num_parallel_calls=1
    )
    assert not invalid
    # TODO: add more cases.


def test_verify_features_scalar_labels(csv_of_volumes):  # noqa: F811
    files = io.read_csv(csv_of_volumes, skip_header=False)
    # Int labels.
    files = [(x, 0) for (x, _) in files]
    invalid = io.verify_features_labels(
        files, volume_shape=(8, 8, 8), num_parallel_calls=1
    )
    assert not invalid
    invalid = io.verify_features_labels(
        files, volume_shape=(12, 12, 8), num_parallel_calls=1
    )
    assert all(invalid)
    # Float labels.
    files = [(x, 1.0) for (x, _) in files]
    invalid = io.verify_features_labels(
        files, volume_shape=(8, 8, 8), num_parallel_calls=1
    )
    assert not invalid
    invalid = io.verify_features_labels(
        files, volume_shape=(12, 12, 8), num_parallel_calls=1
    )
    assert all(invalid)


@pytest.mark.parametrize("filesys", [None, LocalFileSystem()])
def test_is_gzipped(tmp_path, filesys):
    filename = str(tmp_path / "test.gz")
    with gzip.GzipFile(filename, "w") as f:
        f.write("i'm more than a test!".encode())
    assert io._is_gzipped(filename, filesys=filesys)

    with open(filename, "w") as f:
        f.write("i'm just a test...")
    assert not io._is_gzipped(filename, filesys=filesys)



import numpy as np
import tensorflow as tf
import tensorflow.compat.v2 as tf2
from tensorflow.python.ops import nn_impl
import tensorflow_probability as tfp
from tensorflow_probability.python import util as tfp_util
from tensorflow_probability.python.distributions import (
    deterministic as deterministic_lib,
)
from tensorflow_probability.python.distributions import independent as independent_lib
from tensorflow_probability.python.distributions import normal as normal_lib

tfd = tfp.distributions


def default_loc_scale_fn(
    is_singular=True,
    loc_initializer=tf.keras.initializers.he_normal(),
    untransformed_scale_initializer=tf.constant_initializer(0.0001),
    loc_regularizer=None,
    untransformed_scale_regularizer=None,
    loc_constraint=None,
    untransformed_scale_constraint=None,
    weightnorm=False,
):
    """Function for `mean`, `std`, 'weightnorm' tf.variables.

    This function produces a closure which produces `mean`, `std`, 'weightnorm'
    using`tf.get_variable`. The closure accepts the following arguments:
    dtype: Type of parameter's event.
    shape: Python `list`-like representing the parameter's event shape.
    name: Python `str` name prepended to any created (or existing)
     `tf.Variable`s.
    trainable: Python `bool` indicating all created `tf.Variable`s should be
     added to the graph collection `GraphKeys.TRAINABLE_VARIABLES`.
    add_variable_fn: `tf.get_variable`-like `callable` used to create (or
     access existing) `tf.Variable`s.

    Parameters
    ----------
        is_singular: Python `bool` indicating if `scale is None`. Default: `False`.
        loc_initializer: Initializer function for the `loc` parameters.
          The default is `tf.random_normal_initializer(mean=0., stddev=0.1)`.
        untransformed_scale_initializer: Initializer function for the `scale`
          parameters. Default value: `tf.random_normal_initializer(mean=-3.,
          stddev=0.1)`. This implies the softplus transformed result is initialized
          near `0`. It allows a `Normal` distribution with `scale` parameter set to
          this value to approximately act like a point mass.
        loc_regularizer: Regularizer function for the `loc` parameters.
          The default (`None`) is to use the `tf.get_variable` default.
        untransformed_scale_regularizer: Regularizer function for the `scale`
          parameters. The default (`None`) is to use the `tf.get_variable` default.
        loc_constraint: An optional projection function to be applied to the
          loc after being updated by an `Optimizer`. The function must take as input
          the unprojected variable and must return the projected variable (which
          must have the same shape). Constraints are not safe to use when doing
          asynchronous distributed training.
          The default (`None`) is to use the `tf.get_variable` default.
        untransformed_scale_constraint: An optional projection function to be
          applied to the `scale` parameters after being updated by an `Optimizer`
          (e.g. used to implement norm constraints or value constraints). The
          function must take as input the unprojected variable and must return the
          projected variable (which must have the same shape). Constraints are not
          safe to use when doing asynchronous distributed training. The default
          (`None`) is to use the `tf.get_variable` default.
        weightnorm: An optional (boolean) to activate weightnorm for the mean
        kernal.

    Returns
    ----------
        default_loc_scale_fn: Python `callable` which instantiates `loc`, `scale`
        parameters from args: `dtype, shape, name, trainable, add_variable_fn`.

    """

    def _fn(dtype, shape, name, trainable, add_variable_fn):
        """Creates `loc`, `scale` and weightnorm parameters."""
        loc = add_variable_fn(
            name=name + "_loc",
            shape=shape,
            initializer=loc_initializer,
            regularizer=loc_regularizer,
            constraint=loc_constraint,
            dtype=dtype,
            trainable=trainable,
        )
        if weightnorm:
            g = add_variable_fn(
                name=name + "_wn",
                shape=shape,
                initializer=tf.constant_initializer(1.4142),
                constraint=loc_constraint,
                regularizer=loc_regularizer,
                dtype=dtype,
                trainable=trainable,
            )
            loc_wn = tfp_util.DeferredTensor(
                loc, lambda x: (tf.multiply(nn_impl.l2_normalize(x), g))
            )
        # loc = tfp_util.DeferredTensor(loc, lambda x: (nn_impl.l2_normalize(x)))
        if is_singular:
            if weightnorm:
                return loc_wn, None
            else:
                return loc, None

        untransformed_scale = add_variable_fn(
            name=name + "_untransformed_scale",
            shape=shape,
            initializer=untransformed_scale_initializer,
            regularizer=untransformed_scale_regularizer,
            constraint=untransformed_scale_constraint,
            dtype=dtype,
            trainable=trainable,
        )
        scale = tfp_util.DeferredTensor(
            untransformed_scale,
            lambda x: (np.finfo(dtype.as_numpy_dtype).eps + tf.nn.softplus(x)),
        )
        if weightnorm:
            return loc_wn, scale
        else:
            return loc, scale

    return _fn


def default_mean_field_normal_fn(
    is_singular=False,
    loc_initializer=tf.keras.initializers.he_normal(),
    untransformed_scale_initializer=tf.constant_initializer(0.0001),
    loc_regularizer=None,  # tf.keras.regularizers.l2(), #None
    untransformed_scale_regularizer=None,  # tf.keras.regularizers.l2(), #None
    loc_constraint=None,  # tf.keras.constraints.UnitNorm(axis = [0, 1, 2,3]),
    untransformed_scale_constraint=None,
    weightnorm=False,
):
    """Function for deterministic/variational layers.

    This function produces a closure which produces `tfd.Normal`or
    tfd.Deterministic parameterized by a `loc` and `scale` each created
    using `tf.get_variable`.

    Parameters
    ----------
        is_singular: Python `bool` if `True`, forces the special case limit of
          `scale->0`, i.e., a `Deterministic` distribution. and if set False
          put a variational layer.
        loc_initializer: Initializer function for the `loc` parameters.
          The default is `tf.random_normal_initializer(mean=0., stddev=0.1)`.
        untransformed_scale_initializer: Initializer function for the `scale`
          parameters. Default value: `tf.random_normal_initializer(mean=-3.,
          stddev=0.1)`. This implies the softplus transformed result is initialized
          near `0`. It allows a `Normal` distribution with `scale` parameter set to
          this value to approximately act like a point mass.
        loc_regularizer: Regularizer function for the `loc` parameters.
        untransformed_scale_regularizer: Regularizer function for the `scale`
          parameters.
        loc_constraint: An optional projection function to be applied to the
          loc after being updated by an `Optimizer`. The function must take as input
          the unprojected variable and must return the projected variable (which
          must have the same shape). Constraints are not safe to use when doing
          asynchronous distributed training.
        untransformed_scale_constraint: An optional projection
        function to behttps://github.com/FedericoVasile1/Project8
          applied to the `scale` parameters after being updated by an `Optimizer`
          (e.g. used to implement norm constraints or value constraints). The
          function must take as input the unprojected variable and must return the
          projected variable (which must have the same shape). Constraints are not
          safe to use when doing asynchronous distributed training.

    Returns
    ----------
        default_mean_field_normal_fn: Python `callable` which instantiates layers.
    """
    loc_scale_fn = default_loc_scale_fn(
        is_singular=is_singular,
        loc_initializer=loc_initializer,
        untransformed_scale_initializer=untransformed_scale_initializer,
        loc_regularizer=loc_regularizer,
        untransformed_scale_regularizer=untransformed_scale_regularizer,
        loc_constraint=loc_constraint,
        untransformed_scale_constraint=untransformed_scale_constraint,
        weightnorm=weightnorm,
    )

    def _fn(dtype, shape, name, trainable, add_variable_fn):
        loc, scale = loc_scale_fn(dtype, shape, name, trainable, add_variable_fn)
        if scale is None:
            dist = deterministic_lib.Deterministic(loc=loc)
        else:
            dist = normal_lib.Normal(loc=loc, scale=scale)
        batch_ndims = tf2.size(dist.batch_shape_tensor())
        return independent_lib.Independent(dist, reinterpreted_batch_ndims=batch_ndims)

    return _fn


def divergence_fn_bayesian(prior_std, examples_per_epoch):
    """Computes KLD loss.

    Computes KLD function for ELBO loss with
    examples per epochs as scaling parameter.

    Parameters
    ----------
        prior_std: int, scale parameter for the priors
        examples_per_epoch: int, scaling number (batchsize)

    Returns
    ----------
        Scaled KLD loss.
    """

    def divergence_fn(q, p, _):
        log_probs = tfd.LogNormal(0.0, prior_std).log_prob(p.stddev())
        out = tfd.kl_divergence(q, p) - tf.reduce_sum(log_probs)
        return out / examples_per_epoch

    return divergence_fn


def prior_fn_for_bayesian(init_scale_mean=-1, init_scale_std=0.1):
    """Set priors for the variational layers, possibly trainable.

    Parameters
    ----------
        iniy_scale_mean: int, mean initialized value
        init_scale_std: int, scale initiatlization

    Returns
    ----------
        prior_fn: Python `callable` which instantiates `loc`, `scale`.
    """

    def prior_fn(dtype, shape, name, _, add_variable_fn):
        untransformed_scale = add_variable_fn(
            name=name + "_untransformed_scale",
            shape=(1,),
            initializer=tf.compat.v1.initializers.random_normal(
                mean=init_scale_mean, stddev=init_scale_std
            ),
            dtype=dtype,
            trainable=True,
        )
        loc = add_variable_fn(
            name=name + "_loc",
            initializer=tf.keras.initializers.Zeros(),
            shape=shape,
            dtype=dtype,
            trainable=True,
        )
        scale = 1e-4 + tf.nn.softplus(untransformed_scale)
        dist = tfd.Normal(loc=loc, scale=scale)
        batch_ndims = tf.size(input=dist.batch_shape_tensor())
        return tfd.Independent(dist, reinterpreted_batch_ndims=batch_ndims)

    return prior_fn


def normal_prior(prior_std=1.0):
    """Sets normal distributions prior.

    Parameters
    ----------
        prior_std: int, scale parameter, default setting is 1.0.

    Returns
    ----------
        prior_fn: Python `callable`normal distribution.
    """

    def prior_fn(dtype, shape, name, trainable, add_variable_fn):
        dist = tfd.Normal(
            loc=tf.zeros(shape, dtype), scale=dtype.as_numpy_dtype((prior_std))
        )
        batch_ndims = tf.size(input=dist.batch_shape_tensor())
        return tfd.Independent(dist, reinterpreted_batch_ndims=batch_ndims)

    return prior_fn



"""Methods for creating `tf.data.Dataset` objects."""

import math

import fsspec
import numpy as np
import tensorflow as tf

from .io import _is_gzipped
from .tfrecord import parse_example_fn
from .volume import (
    apply_random_transform,
    apply_random_transform_scalar_labels,
    binarize,
    replace,
    standardize,
    to_blocks,
)

AUTOTUNE = tf.data.experimental.AUTOTUNE


def tfrecord_dataset(
    file_pattern,
    volume_shape,
    shuffle,
    scalar_label,
    compressed=True,
    num_parallel_calls=AUTOTUNE,
):
    """Return `tf.data.Dataset` from TFRecord files."""
    dataset = tf.data.Dataset.list_files(file_pattern, shuffle=shuffle)
    # Read each of these files as a TFRecordDataset.
    # Assume all files have same compression type as the first file.
    compression_type = "GZIP" if compressed else None
    cycle_length = 1 if num_parallel_calls is None else num_parallel_calls
    parse_fn = parse_example_fn(volume_shape=volume_shape, scalar_label=scalar_label)

    if not shuffle:
        # Determine examples_per_shard from the first TFRecord shard
        # Then set block_length to equal the number of examples per shard
        # so that the interleave method does not inadvertently shuffle data.
        first_shard = (
            dataset.take(1)
            .flat_map(
                lambda x: tf.data.TFRecordDataset(x, compression_type=compression_type)
            )
            .map(map_func=parse_fn, num_parallel_calls=num_parallel_calls)
        )
        block_length = len([0 for _ in first_shard])
    else:
        # If the dataset is being shuffled, then we don't care if interleave
        # further shuffles that data even further
        block_length = None

    dataset = dataset.interleave(
        map_func=lambda x: tf.data.TFRecordDataset(
            x, compression_type=compression_type
        ),
        cycle_length=cycle_length,
        block_length=block_length,
        num_parallel_calls=num_parallel_calls,
    )
    dataset = dataset.map(map_func=parse_fn, num_parallel_calls=num_parallel_calls)
    return dataset


def get_dataset(
    file_pattern,
    n_classes,
    batch_size,
    volume_shape,
    scalar_label=False,
    block_shape=None,
    n_epochs=None,
    mapping=None,
    augment=False,
    normalizer=standardize,
    shuffle_buffer_size=None,
    num_parallel_calls=AUTOTUNE,
):
    """Return `tf.data.Dataset` that preprocesses data for training or prediction.

    Labels are preprocessed for binary or multiclass segmentation according to
    `n_classes`.

    Parameters
    ----------
    file_pattern: str, expression that can be globbed to get TFRecords files
        for this dataset. For example 'data/training_*.tfrecords'.
    n_classes: int, number of classes to segment. Values of 1 and 2 indicate
        binary segmentation (foreground vs background), and values greater than
        2 indicate multiclass segmentation.
    batch_size: int, number of elements per batch.
    volume_shape: tuple of at least length 3, the shape of every volume in the TFRecords
        files. Every volume must have the same shape.
    scalar_label: boolean, if `True`, labels are scalars.
    block_shape: tuple of at least length 3, the shape of the non-overlapping sub-volumes
        to take from the full volumes. If None, do not separate the full volumes
        into sub-volumes. Separating into non-overlapping sub-volumes is useful
        (sometimes even necessary) to overcome memory limitations depending on
        the number of model parameters.
    n_epochs: int, number of epochs for the dataset to repeat. If None, the
        dataset will be repeated indefinitely.
    mapping: dict, mapping to replace label values. Values equal to a key in
        the mapping are replaced with the corresponding values in the mapping.
        Values not in `mapping.keys()` are replaced with zeros.
    augment: boolean, if true, apply random rigid transformations to the
        features and labels. The rigid transformations are applied to the full
        volumes.
    normalizer: callable, applies this normalization function when creating the
        dataset. to maintain compatibility with prior nobrainer release, this is
        set to standardize by default.
    shuffle_buffer_size: int, buffer of full volumes to shuffle. If this is not
        None, then the list of files found by 'file_pattern' is also shuffled
        at every iteration.
    num_parallel_calls: int, number of parallel calls to make for data loading
        and processing.

    Returns
    -------
    `tf.data.Dataset` of features and labels. If block_shape is not None, the
    shape of features is `(batch_size, *block_shape, 1)` and the shape of labels
    is `(batch_size, *block_shape, n_classes)`. If block_shape is None, then
    the shape of features is `(batch_size, *volume_shape, 1)` and the shape of
    labels is `(batch_size, *volume_shape, n_classes)`. If `scalar_label` is `True,
    the shape of labels is always `(batch_size,)`.
    """
    fs, _, _ = fsspec.get_fs_token_paths(file_pattern)
    files = fs.glob(file_pattern)
    if not files:
        raise ValueError("no files found for pattern '{}'".format(file_pattern))

    # Create dataset of all TFRecord files. After this point, the dataset will have
    # two value per iteration: (feature, label).
    shuffle = bool(shuffle_buffer_size)
    compressed = _is_gzipped(files[0], filesys=fs)
    dataset = tfrecord_dataset(
        file_pattern=file_pattern,
        volume_shape=volume_shape,
        shuffle=shuffle,
        scalar_label=scalar_label,
        compressed=compressed,
        num_parallel_calls=num_parallel_calls,
    )

    if normalizer is not None:
        # Standard-score the features.
        dataset = dataset.map(lambda x, y: (normalizer(x), y))

    # Augment examples if requested.
    if augment:
        if not scalar_label:
            dataset = dataset.map(
                lambda x, y: tf.cond(
                    tf.random.uniform((1,)) > 0.5,
                    true_fn=lambda: apply_random_transform(x, y),
                    false_fn=lambda: (x, y),
                ),
                num_parallel_calls=num_parallel_calls,
            )
        else:
            dataset = dataset.map(
                lambda x, y: tf.cond(
                    tf.random.uniform((1,)) > 0.5,
                    true_fn=lambda: apply_random_transform_scalar_labels(x, y),
                    false_fn=lambda: (x, y),
                ),
                num_parallel_calls=num_parallel_calls,
            )

    # Separate into blocks, if requested.
    if block_shape is not None:
        if not scalar_label:
            dataset = dataset.map(
                lambda x, y: (to_blocks(x, block_shape), to_blocks(y, block_shape)),
                num_parallel_calls=num_parallel_calls,
            )
            # This step is necessary because separating into blocks adds a dimension.
            dataset = dataset.unbatch()
        if scalar_label:

            def _f(x, y):
                x = to_blocks(x, block_shape)
                n_blocks = x.shape[0]
                y = tf.repeat(y, n_blocks)
                return (x, y)

            dataset = dataset.map(_f, num_parallel_calls=num_parallel_calls)
            # This step is necessary because separating into blocks adds a dimension.
            dataset = dataset.unbatch()
    else:
        if scalar_label:
            dataset = dataset.map(lambda x, y: (x, tf.squeeze(y)))

    # Binarize or replace labels according to mapping.
    if not scalar_label:
        if n_classes < 1:
            raise ValueError("n_classes must be > 0.")
        elif n_classes == 1:
            dataset = dataset.map(lambda x, y: (x, tf.expand_dims(binarize(y), -1)))
        elif n_classes == 2:
            dataset = dataset.map(lambda x, y: (x, tf.one_hot(binarize(y), n_classes)))
        elif n_classes > 2:
            if mapping is not None:
                dataset = dataset.map(lambda x, y: (x, replace(y, mapping=mapping)))
            dataset = dataset.map(lambda x, y: (x, tf.one_hot(y, n_classes)))

    # If volume_shape is only three dims, add grayscale channel to features.
    # Otherwise, assume that the channels are already in the features.
    if len(volume_shape) == 3:
        dataset = dataset.map(lambda x, y: (tf.expand_dims(x, -1), y))

    # Prefetch data to overlap data production with data consumption. The
    # TensorFlow documentation suggests prefetching `batch_size` elements.
    dataset = dataset.prefetch(buffer_size=batch_size)

    # Batch the dataset, so each iteration gives `batch_size` elements. We drop
    # the remainder so that when training on multiple GPUs, the batch will
    # always be evenly divisible by the number of GPUs. Otherwise, the last
    # batch might have fewer than `batch_size` elements and will cause errors.
    if batch_size is not None:
        dataset = dataset.batch(batch_size=batch_size, drop_remainder=True)

    # Optionally shuffle. We also optionally shuffle the list of files.
    # The TensorFlow recommend shuffling and then repeating.
    if shuffle_buffer_size:
        dataset = dataset.shuffle(buffer_size=shuffle_buffer_size)

    # Repeat the dataset for n_epochs. If n_epochs is None, then repeat
    # indefinitely. If n_epochs is 1, then the dataset will only be iterated
    # through once.
    dataset = dataset.repeat(n_epochs)

    return dataset


def get_steps_per_epoch(n_volumes, volume_shape, block_shape, batch_size):
    def get_n(a, k):
        return (a - k) / k + 1

    n_blocks = tuple(get_n(aa, kk) for aa, kk in zip(volume_shape, block_shape))

    for n in n_blocks:
        if not n.is_integer() or n < 1:
            raise ValueError(
                "cannot create non-overlapping blocks with the given parameters."
            )
    n_blocks_per_volume = np.prod(n_blocks).astype(int)

    steps = n_blocks_per_volume * n_volumes / batch_size
    steps = math.ceil(steps)
    return steps



from . import _version

__version__ = _version.get_versions()["version"]

from . import (  # noqa: F401
    dataset,
    io,
    layers,
    losses,
    metrics,
    models,
    prediction,
    training,
    transform,
    utils,
    volume,
)



"""Model definition for ProgressiveGAN.
"""

import tensorflow as tf
from tensorflow.keras import layers, models

from ..volume import adjust_dynamic_range as _adjust_dynamic_range


def progressivegan(
    latent_size,
    label_size=0,
    num_channels=1,
    dimensionality=3,
    g_fmap_base=8192,
    d_fmap_base=8192,
    g_fmap_max=256,
    d_fmap_max=256,
):
    """Instantiate ProgressiveGAN Architecture.

    Parameters
    ----------
    latent_size: int, size of the latent space to use for generating images
    label_size: int, number of labels for conditional image synthesis (WIP)
    num_channels: int, number of channels in the generated images
    dimensionality: int, one of [2, 3], number of dimensions in the image
    g_fmap_base: int, parameter to determine width of generator
    d_fmap_base: int, parameter to determine width of discriminator
    g_fmap_max: int, parameter to determine width of generator
    d_fmap_max: int, parameter to determine width of discriminator
    name: str, name to give to the resulting model object.

    Returns
    -------
    Generator and Discriminator
    """

    generator = Generator(
        latent_size,
        label_size=label_size,
        num_channels=num_channels,
        dimensionality=dimensionality,
        fmap_base=g_fmap_base,
        fmap_max=g_fmap_max,
    )

    discriminator = Discriminator(
        label_size=label_size,
        num_channels=num_channels,
        dimensionality=dimensionality,
        fmap_base=d_fmap_base,
        fmap_max=d_fmap_max,
    )

    return generator, discriminator


class Generator(tf.keras.Model):
    def __init__(
        self,
        latent_size,
        label_size=0,
        num_channels=1,
        fmap_base=8192,
        fmap_max=256,
        dimensionality=3,
    ):
        super(Generator, self).__init__()

        self.latent_size = latent_size
        self.label_size = label_size

        self.fmap_base = fmap_base
        self.fmap_max = fmap_max
        self.num_channels = num_channels
        self.dimensionality = dimensionality

        self.Conv = getattr(layers, "Conv{}D".format(self.dimensionality))
        self.ConvTranspose = getattr(
            layers, "Conv{}DTranspose".format(self.dimensionality)
        )
        self.Upsampling = getattr(layers, "UpSampling{}D".format(self.dimensionality))

        self.current_resolution = 2
        self.highest_resolution_block = self._make_generator_block(
            self._nf(self.current_resolution),
            name="g_block_{}".format(self.current_resolution),
        )

        self.resolution_blocks = []

        self.base_dense = tf.keras.layers.Dense(
            units=self._nf(1) * (2 ** self.dimensionality)
        )
        self.HeadConv1 = self.Conv(filters=self.num_channels, kernel_size=1)
        self.HeadConv2 = self.Conv(filters=self.num_channels, kernel_size=1)

        self.build([(None, latent_size), (1,)])

    def _pixel_norm(self, epsilon=1e-8):
        """
        Pixelwise normalization
        """
        return layers.Lambda(
            lambda x: x
            * tf.math.rsqrt(
                tf.reduce_mean(tf.square(x), axis=-1, keepdims=True) + epsilon
            )
        )
        # return layers.BatchNormalization(axis=-1)

    def _weighted_sum(self):
        """
        Weighted interpolation helper for fading in new layers as its progressively added
        """
        return layers.Lambda(
            lambda inputs: (1 - inputs[2]) * inputs[0] + (inputs[2]) * inputs[1]
        )

    def _nf(self, stage):
        """
        Computes number of filters for a conv layer
        """
        return min(int(self.fmap_base / (2.0 ** (stage))), self.fmap_max)

    def generator_base(self, latents):
        """
        Creates the base of the generator with inputs of size latent_size+label_size
        and the alpha value to determine level of interpolation
        """

        # Latents stage
        x = self._pixel_norm()(latents)
        x = self.base_dense(x)
        x = layers.Reshape([2] * self.dimensionality + [self._nf(1)])(x)

        return x

    def _make_generator_block(self, nf, kernel_size=4, name=""):
        """
        Creates a generator block
        """

        block_layers = []

        # block_layers.append(self.ConvTranspose(nf, kernel_size=3,
        # strides=2, padding='same'))
        block_layers.append(self.Upsampling())
        # block_layers.append(self.Conv(filters=nf, kernel_size=kernel_size,
        # strides=1, padding='same'))
        # block_layers.append(layers.Activation(tf.nn.leaky_relu))
        # block_layers.append(self._pixel_norm())

        block_layers.append(
            self.Conv(filters=nf, kernel_size=kernel_size, strides=1, padding="same")
        )
        block_layers.append(layers.Activation(tf.nn.leaky_relu))
        block_layers.append(self._pixel_norm())

        return models.Sequential(block_layers, name=name)

    def generator_head(self, x, y, alpha):

        x = self.Upsampling()(x)
        x = self.HeadConv1(x)

        y = self.HeadConv2(y)

        output = self._weighted_sum()([x, y, alpha])
        output = layers.Activation("tanh")(output)

        return output

    def add_resolution(self):

        self.current_resolution += 1
        self.resolution_blocks.append(self.highest_resolution_block)
        self.highest_resolution_block = self._make_generator_block(
            self._nf(self.current_resolution),
            name="g_block_{}".format(self.current_resolution),
        )

        self.HeadConv1 = self.Conv(filters=self.num_channels, kernel_size=1)
        self.HeadConv2 = self.Conv(filters=self.num_channels, kernel_size=1)

        self.build([(None, self.latent_size), (1,)])

    def call(self, inputs):
        latents, alpha = inputs
        x = self.generator_base(latents)

        for g_block in self.resolution_blocks:
            x = g_block(x)

        y = self.highest_resolution_block(x)
        return self.generator_head(x, y, alpha)

    def generate(self, latents):
        alpha = tf.constant([1.0], tf.float32)
        image = self.call([latents, alpha])
        image = _adjust_dynamic_range(image, [-1, 1], [0, 255])
        return {"generated": image}

    def save(self, filepath, **kwargs):
        # force retrace of tf.function to track new and deleted variables
        super().save(
            filepath,
            signatures=tf.function(
                self.generate,
                input_signature=(
                    tf.TensorSpec(shape=[None, self.latent_size], dtype=tf.float32),
                ),
            ),
            **kwargs
        )


class Discriminator(tf.keras.Model):
    """
    Progressive Discriminator
    """

    def __init__(
        self,
        label_size=0,
        num_channels=1,
        fmap_base=8192,
        fmap_max=512,
        dimensionality=3,
    ):
        super(Discriminator, self).__init__()

        self.label_size = label_size

        self.fmap_base = fmap_base
        self.fmap_max = fmap_max
        self.num_channels = num_channels
        self.dimensionality = dimensionality

        self.current_resolution = 2

        self.Conv = getattr(layers, "Conv{}D".format(self.dimensionality))
        self.AveragePooling = getattr(
            layers, "AveragePooling{}D".format(self.dimensionality)
        )

        self.highest_resolution_block = self._make_discriminator_block(
            self._nf(self.current_resolution - 1),
            name="d_block_{}".format(self.current_resolution),
        )
        self.resolution_blocks = []

        self.BaseConv = self.Conv(
            filters=self._nf(self.current_resolution - 1), kernel_size=1, padding="same"
        )
        self.FadeConv = self.Conv(
            filters=self._nf(self.current_resolution), kernel_size=1, padding="same"
        )
        self.HeadDense1 = tf.keras.layers.Dense(units=self._nf(1))
        self.HeadDense2 = tf.keras.layers.Dense(units=1 + self.label_size)

        images_shape = (
            (None,)
            + (int(2.0 ** self.current_resolution),) * self.dimensionality
            + (self.num_channels,)
        )
        alpha_shape = (1,)

        self.build([images_shape, alpha_shape])

    def _weighted_sum(self):
        """
        Weighted interpolation helper for fading in new layers as its progressively added
        """
        return layers.Lambda(
            lambda inputs: (1 - inputs[2]) * inputs[0] + (inputs[2]) * inputs[1]
        )

    def _nf(self, stage):
        """
        Computes number of filters for a conv layer
        """
        return min(int(self.fmap_base / (2.0 ** (stage))), self.fmap_max)

    def discriminator_base(self, x, y, alpha):

        x = self.AveragePooling()(x)
        x = self.BaseConv(x)

        lerp_input = self._weighted_sum()([x, y, alpha])

        return lerp_input

    def _make_discriminator_block(self, nf, kernel_size=4, name=""):
        """
        Creates a discriminator block
        """

        block_layers = []

        # block_layers.append(self.Conv(filters=nf, kernel_size=kernel_size,
        # strides=1, padding='same'))
        # block_layers.append(layers.Activation(tf.nn.leaky_relu))

        block_layers.append(
            self.Conv(filters=nf, kernel_size=kernel_size, strides=2, padding="same")
        )
        block_layers.append(layers.Activation(tf.nn.leaky_relu))

        return models.Sequential(block_layers, name=name)

    def discriminator_head(self, x):
        """
        Creates the end of the disciminator with inputs of image dimensions and the
        alpha value to determine level of interpolation. The output node gives
        the prediction of fakeness of the input image.
        """
        x = layers.Flatten()(x)
        x = self.HeadDense1(x)
        x = layers.Activation(tf.nn.leaky_relu)(x)

        output = self.HeadDense2(x)

        score_output = layers.Lambda(lambda x: x[..., 0])(output)
        label_output = layers.Lambda(lambda x: x[..., 1:])(output)

        return score_output, label_output

    def add_resolution(self):
        """
        Adds a resolution step to the trainable discriminator by creating a
        discriminator block and inserting at the start of the existing discriminator
        to increase the resolution of the images discimrinated by a factor of 2.
        The new layers are faded in and controlled by the alpha parameter.
        """

        self.current_resolution += 1

        self.resolution_blocks.append(self.highest_resolution_block)
        self.highest_resolution_block = self._make_discriminator_block(
            self._nf(self.current_resolution - 1),
            name="d_block_{}".format(self.current_resolution),
        )

        self.BaseConv = self.Conv(
            filters=self._nf(self.current_resolution - 1), kernel_size=1, padding="same"
        )
        self.FadeConv = self.Conv(
            filters=self._nf(self.current_resolution), kernel_size=1, padding="same"
        )

        images_shape = (
            (None,)
            + (int(2.0 ** self.current_resolution),) * self.dimensionality
            + (self.num_channels,)
        )
        alpha_shape = (1,)

        self.build([images_shape, alpha_shape])

    def call(self, inputs):

        images, alpha = inputs

        # To bring to the right number of filters
        x = self.FadeConv(images)
        y = self.highest_resolution_block(x)
        x = self.discriminator_base(images, y, alpha)

        for d_block in self.resolution_blocks[::-1]:
            x = d_block(x)

        return self.discriminator_head(x)



"""Custom padding layers for nobrainer."""

import tensorflow as tf
from tensorflow.keras import layers


class ZeroPadding3DChannels(layers.Layer):
    """Pad the last dimension of a 5D tensor symmetrically with zeros.

    This is meant for 3D convolutions, where tensors are 5D.
    """

    def __init__(self, padding, **kwds):
        self.padding = padding
        # batch, x, y, z, channels
        self._paddings = [[0, 0], [0, 0], [0, 0], [0, 0], [self.padding, self.padding]]
        super(ZeroPadding3DChannels, self).__init__(**kwds)

    def call(self, x):
        return tf.pad(x, paddings=self._paddings, mode="CONSTANT", constant_values=0)



from .autoencoder import autoencoder
from .dcgan import dcgan
from .highresnet import highresnet
from .meshnet import meshnet
from .progressivegan import progressivegan
from .unet import unet


def get(name):
    """Return callable that creates a particular `tf.keras.Model`.

    Parameters
    ----------
    name: str, the name of the model (case-insensitive).

    Returns
    -------
    Callable, which instantiates a `tf.keras.Model` object.
    """
    if not isinstance(name, str):
        raise ValueError("Model name must be a string.")

    models = {
        "highresnet": highresnet,
        "meshnet": meshnet,
        "unet": unet,
        "autoencoder": autoencoder,
        "progressivegan": progressivegan,
        "dcgan": dcgan,
    }

    try:
        return models[name.lower()]
    except KeyError:
        avail = ", ".join(models.keys())
        raise ValueError(
            "Unknown model: '{}'. Available models are {}.".format(name, avail)
        )



import glob
import os.path as op
import shutil
import tempfile

import nibabel as nib
import numpy as np
from numpy.testing import assert_array_equal
import pytest

from .. import dataset, io, tfrecord, utils


@pytest.fixture(scope="session")
def tmp_data_filepaths():
    temp_dir = tempfile.mkdtemp()
    csv_of_filepaths = utils.get_data(cache_dir=temp_dir)
    filepaths = io.read_csv(csv_of_filepaths)
    yield filepaths
    shutil.rmtree(temp_dir)


def write_tfrecs(filepaths, outdir, examples_per_shard):
    tfrecord.write(
        features_labels=filepaths,
        filename_template=op.join(outdir, "data_shard-{shard:03d}.tfrec"),
        examples_per_shard=examples_per_shard,
    )
    return op.join(outdir, "data_shard-*.tfrec")


@pytest.mark.parametrize("examples_per_shard", [1, 3])
@pytest.mark.parametrize("batch_size", [1, 2])
@pytest.mark.parametrize("num_parallel_calls", [1, 2])
def test_get_dataset_maintains_order(
    tmp_data_filepaths, examples_per_shard, batch_size, num_parallel_calls
):
    filepaths = [(x, i) for i, (x, _) in enumerate(tmp_data_filepaths)]
    temp_dir = tempfile.mkdtemp()
    file_pattern = write_tfrecs(
        filepaths, temp_dir, examples_per_shard=examples_per_shard
    )
    volume_shape = (256, 256, 256)
    dset = dataset.get_dataset(
        file_pattern=file_pattern,
        n_classes=1,
        batch_size=batch_size,
        volume_shape=volume_shape,
        scalar_label=True,
        n_epochs=1,
        num_parallel_calls=num_parallel_calls,
    )
    y_orig = np.array([y for _, y in filepaths])
    y_from_dset = (
        np.concatenate([y for _, y in dset.as_numpy_iterator()]).flatten().astype(int)
    )
    assert_array_equal(y_orig, y_from_dset)
    shutil.rmtree(temp_dir)


def create_dummy_niftis(shape, n_volumes, outdir):
    array_data = np.zeros(shape, dtype=np.int16)
    affine = np.diag([1, 2, 3, 1])
    array_img = nib.Nifti1Image(array_data, affine)
    for volume in range(n_volumes):
        nib.save(array_img, op.join(outdir, f"image_{volume}.nii.gz"))

    return glob.glob(op.join(outdir, "image_*.nii.gz"))


def test_get_dataset_errors():
    temp_dir = tempfile.mkdtemp()
    file_pattern = op.join(temp_dir, "does_not_exist-*.tfrec")
    with pytest.raises(ValueError):
        dataset.get_dataset(
            file_pattern=file_pattern,
            n_classes=1,
            batch_size=1,
            volume_shape=(256, 256, 256),
        )


@pytest.mark.parametrize("batch_size", [1, 2])
@pytest.mark.parametrize("examples_per_shard", [1, 3])
@pytest.mark.parametrize("volume_shape", [(64, 64, 64), (64, 64, 64, 3)])
@pytest.mark.parametrize("num_parallel_calls", [1, 2])
def test_get_dataset_shapes(
    volume_shape, examples_per_shard, batch_size, num_parallel_calls
):
    temp_dir = tempfile.mkdtemp()
    nifti_paths = create_dummy_niftis(volume_shape, 10, temp_dir)
    filepaths = [(x, i) for i, x in enumerate(nifti_paths)]
    file_pattern = write_tfrecs(
        filepaths, temp_dir, examples_per_shard=examples_per_shard
    )
    dset = dataset.get_dataset(
        file_pattern=file_pattern,
        n_classes=1,
        batch_size=batch_size,
        volume_shape=volume_shape,
        scalar_label=True,
        n_epochs=1,
        num_parallel_calls=num_parallel_calls,
    )

    output_volume_shape = volume_shape if len(volume_shape) > 3 else volume_shape + (1,)
    output_volume_shape = (batch_size,) + output_volume_shape
    shapes = [x.shape for x, _ in dset.as_numpy_iterator()]
    assert all([_shape == output_volume_shape for _shape in shapes])
    shutil.rmtree(temp_dir)


# TODO: need to implement this soon.
@pytest.mark.xfail
def test_get_dataset():
    assert False


def test_get_steps_per_epoch():
    nsteps = dataset.get_steps_per_epoch(
        n_volumes=1,
        volume_shape=(256, 256, 256),
        block_shape=(64, 64, 64),
        batch_size=1,
    )
    assert nsteps == 64
    nsteps = dataset.get_steps_per_epoch(
        n_volumes=1,
        volume_shape=(256, 256, 256),
        block_shape=(64, 64, 64),
        batch_size=64,
    )
    assert nsteps == 1
    nsteps = dataset.get_steps_per_epoch(
        n_volumes=1,
        volume_shape=(256, 256, 256),
        block_shape=(64, 64, 64),
        batch_size=63,
    )
    assert nsteps == 2
    nsteps = dataset.get_steps_per_epoch(
        n_volumes=10,
        volume_shape=(256, 256, 256),
        block_shape=(128, 128, 128),
        batch_size=4,
    )
    assert nsteps == 20



import numpy as np
import pytest
import tensorflow as tf

from nobrainer.bayesian_utils import default_mean_field_normal_fn

from ..autoencoder import autoencoder
from ..bayesian_vnet import bayesian_vnet
from ..bayesian_vnet_semi import bayesian_vnet_semi
from ..brainsiam import brainsiam
from ..dcgan import dcgan
from ..highresnet import highresnet
from ..meshnet import meshnet
from ..progressivegan import progressivegan
from ..unet import unet
from ..vnet import vnet


def model_test(model_cls, n_classes, input_shape, kwds={}):
    """Tests for models."""
    x = 10 * np.random.random(input_shape)
    y = np.random.choice([True, False], input_shape)

    # Assume every model class has n_classes and input_shape arguments.
    model = model_cls(n_classes=n_classes, input_shape=input_shape[1:], **kwds)
    model.compile(tf.optimizers.Adam(), "binary_crossentropy")
    model.fit(x, y)

    actual_output = model.predict(x)
    assert actual_output.shape == x.shape[:-1] + (n_classes,)


def test_highresnet():
    model_test(highresnet, n_classes=1, input_shape=(1, 32, 32, 32, 1))


def test_meshnet():
    model_test(
        meshnet,
        n_classes=1,
        input_shape=(1, 32, 32, 32, 1),
        kwds={"receptive_field": 37},
    )
    model_test(
        meshnet,
        n_classes=1,
        input_shape=(1, 32, 32, 32, 1),
        kwds={"receptive_field": 67},
    )
    model_test(
        meshnet,
        n_classes=1,
        input_shape=(1, 32, 32, 32, 1),
        kwds={"receptive_field": 129},
    )
    with pytest.raises(ValueError):
        model_test(
            meshnet,
            n_classes=1,
            input_shape=(1, 32, 32, 32, 1),
            kwds={"receptive_field": 50},
        )


def test_unet():
    model_test(unet, n_classes=1, input_shape=(1, 32, 32, 32, 1))


def test_autoencoder():
    """Special test for autoencoder."""

    input_shape = (1, 32, 32, 32, 1)
    x = 10 * np.random.random(input_shape)

    model = autoencoder(input_shape[1:], encoding_dim=128, n_base_filters=32)
    model.compile(tf.optimizers.Adam(), "mse")
    model.fit(x, x)

    actual_output = model.predict(x)
    assert actual_output.shape == x.shape


def test_brainsiam():
    """Testing the encoder-projector and predictor structures of the brainsiam architecture"""
    input_shape = (1, 32, 32, 32, 1)
    x = 10 * np.random.random(input_shape)

    n_classes = 1
    weight_decay = 0.0005
    projection_dim = 2048
    latent_dim = 512

    encoder, predictor = brainsiam(
        n_classes,
        input_shape=input_shape[1:],
        weight_decay=weight_decay,
        projection_dim=projection_dim,
        latent_dim=latent_dim,
    )

    encoder_output = encoder(x[1:])
    enc_output_shape = encoder_output.get_shape().as_list()

    predictor_out = predictor(encoder_output)
    pred_output_shape = predictor_out.get_shape().as_list()

    assert (
        enc_output_shape[1] == projection_dim
    ), "encoder output shape not the same as projection dim"
    assert (
        pred_output_shape[1] == projection_dim
    ), "predictor output shape not the same as projection dim"


def test_progressivegan():
    """Test for both discriminator and generator of progressive gan"""

    latent_size = 256
    label_size = 2
    g_fmap_base = 1024
    d_fmap_base = 1024
    alpha = 1.0

    generator, discriminator = progressivegan(
        latent_size,
        label_size=label_size,
        g_fmap_base=g_fmap_base,
        d_fmap_base=d_fmap_base,
    )

    resolutions = [8, 16]

    for res in resolutions:
        generator.add_resolution()
        discriminator.add_resolution()

        latent_input = np.random.random((10, latent_size))
        real_image_input = np.random.random((10, res, res, res, 1))

        fake_images = generator([latent_input, alpha])
        real_pred, real_labels_pred = discriminator([real_image_input, alpha])
        fake_pred, fake_labels_pred = discriminator([fake_images, alpha])

        assert fake_images.shape == real_image_input.shape
        assert real_pred.shape == (real_image_input.shape[0],)
        assert fake_pred.shape == (real_image_input.shape[0],)
        assert real_labels_pred.shape == (real_image_input.shape[0], label_size)
        assert fake_labels_pred.shape == (real_image_input.shape[0], label_size)


def test_dcgan():
    """Special test for dcgan."""

    output_shape = (1, 32, 32, 32, 1)
    z_dim = 32
    z = np.random.random((1, z_dim))

    pred_shape = (1, 8, 8, 8, 1)

    generator, discriminator = dcgan(output_shape[1:], z_dim=z_dim)
    generator.compile(tf.optimizers.Adam(), "mse")
    discriminator.compile(tf.optimizers.Adam(), "mse")

    fake_images = generator.predict(z)
    fake_pred = discriminator.predict(fake_images)

    assert fake_images.shape == output_shape and fake_pred.shape == pred_shape


def test_vnet():
    model_test(vnet, n_classes=1, input_shape=(1, 32, 32, 32, 1))


def model_test_bayesian(model_cls, n_classes, input_shape, kernel_posterior_fn):
    """Tests for models."""
    x = 10 * np.random.random(input_shape)
    y = np.random.choice([True, False], input_shape)

    # Assume every model class has n_classes and input_shape arguments.
    model = model_cls(
        n_classes=n_classes,
        input_shape=input_shape[1:],
        kernel_posterior_fn=kernel_posterior_fn,
    )
    model.compile(tf.optimizers.Adam(), "binary_crossentropy")
    model.fit(x, y)

    actual_output = model.predict(x)
    assert actual_output.shape == x.shape[:-1] + (n_classes,)


def test_bayesian_vnet_semi():
    model_test_bayesian(
        bayesian_vnet_semi,
        n_classes=1,
        input_shape=(1, 32, 32, 32, 1),
        kernel_posterior_fn=default_mean_field_normal_fn(weightnorm=True),
    )


def test_bayesian_vnet():
    model_test_bayesian(
        bayesian_vnet,
        n_classes=1,
        input_shape=(1, 32, 32, 32, 1),
        kernel_posterior_fn=default_mean_field_normal_fn(weightnorm=True),
    )



"""training utilities (supports training on single and multiple GPUs)"""

import os

import tensorflow as tf
from tensorflow.python.keras.engine import compile_utils

from .losses import gradient_penalty
from .volume import adjust_dynamic_range as _adjust_dynamic_range


class ProgressiveGANTrainer(tf.keras.Model):
    """Progressive Generative Adversarial Network Trainer.

    Trains discriminator and generator alternatively in an adversarial manner for generation of
    brain MRI images. Uses a progressive method for each resolution and supports the smooth
    transition of new layers, as explained in the reference.

    Parameters
    ----------
    discriminator : tf.keras.Model, Instantiated using nobrainer.models
    generator : tf.keras.Model, Instantiated using nobrainer.models
    gradient_penalty : boolean, Use gradient penalty on discriminator for smooth training.

    References
    ----------
    Progressive Growing of GANs for Improved Quality, Stability, and Variation.
    T. Karras, T. Aila, S. Laine & J. Lehtinen, International Conference on
    Learning Representations. 2018.

    Links
    -----
    [https://research.nvidia.com/sites/default/files/pubs/2017-10_Progressive-Growing-of/karras2018iclr-paper.pdf](https://research.nvidia.com/sites/default/files/pubs/2017-10_Progressive-Growing-of/karras2018iclr-paper.pdf)
    """

    def __init__(self, discriminator, generator, gradient_penalty=False):
        super(ProgressiveGANTrainer, self).__init__()
        self.discriminator = discriminator
        self.generator = generator
        self.gradient_penalty = gradient_penalty
        self.latent_size = generator.latent_size
        self.resolution = 8  # Default resolution
        self.train_step_counter = tf.Variable(
            0.0
        )  # For calculating alpha in transition phase
        self.phase = tf.Variable("resolution")  # For determining whether alpha is 1

    def compile(self, d_optimizer, g_optimizer, g_loss_fn, d_loss_fn):
        super(ProgressiveGANTrainer, self).compile()
        self.d_optimizer = d_optimizer
        self.g_optimizer = g_optimizer

        self.g_loss_fn = compile_utils.LossesContainer(g_loss_fn)
        self.d_loss_fn = compile_utils.LossesContainer(d_loss_fn)

        if self.gradient_penalty:
            self.gradient_penalty_fn = compile_utils.LossesContainer(gradient_penalty)

    def train_step(self, reals):
        if isinstance(reals, tuple):
            reals = reals[0]

        # get batch size dynamically
        batch_size = tf.shape(reals)[0]

        # normalize the real images using minmax to [-1, 1]
        reals = _adjust_dynamic_range(reals, [0.0, 255.0], [-1.0, 1.0])

        # calculate alpha differently for transition and resolution phase
        self.train_step_counter.assign_add(1.0)
        alpha = tf.cond(
            tf.math.equal(self.phase, "transition"),
            lambda: self.train_step_counter / self.steps_per_epoch,
            lambda: tf.constant([1.0]),
        )
        alpha = tf.reshape(alpha, (1,))

        # train discriminator
        latents = tf.random.normal((batch_size, self.latent_size))
        fake_labels = tf.ones((batch_size, 1)) * -1
        real_labels = tf.ones((batch_size, 1))

        with tf.GradientTape() as tape:
            fakes = self.generator([latents, alpha])
            fakes_pred, labels_pred_fake = self.discriminator([fakes, alpha])
            reals_pred, labels_pred_real = self.discriminator([reals, alpha])

            fake_loss = self.d_loss_fn(fake_labels, fakes_pred)
            real_loss = self.d_loss_fn(real_labels, reals_pred)
            d_loss = 0.5 * (fake_loss + real_loss)

            # calculate and add the gradient penalty loss using average samples for discriminator
            if self.gradient_penalty:
                weight_shape = (tf.shape(reals)[0],) + (
                    1,
                    1,
                    1,
                    1,
                )  # broadcasting to right shape
                weight = tf.random.uniform(weight_shape, minval=0, maxval=1)
                average_samples = (weight * reals) + ((1 - weight) * fakes)
                average_pred = self.discriminator(([average_samples, alpha]))
                gradients = tf.gradients(average_pred, average_samples)[0]
                gp_loss = self.gradient_penalty_fn(gradients, reals_pred)
                d_loss += gp_loss

        d_gradients = tape.gradient(d_loss, self.discriminator.trainable_variables)
        self.d_optimizer.apply_gradients(
            zip(d_gradients, self.discriminator.trainable_variables)
        )

        # train generator
        misleading_labels = tf.ones((batch_size, 1))

        latents = tf.random.normal((batch_size, self.latent_size))
        with tf.GradientTape() as tape:
            fakes = self.generator([latents, alpha])
            fakes_pred, labels_pred = self.discriminator([fakes, alpha])

            g_loss = self.g_loss_fn(misleading_labels, fakes_pred)

        g_gradients = tape.gradient(g_loss, self.generator.trainable_variables)
        self.g_optimizer.apply_gradients(
            zip(g_gradients, self.generator.trainable_variables)
        )

        return {"d_loss": d_loss, "g_loss": g_loss}

    def fit(
        self, *args, resolution=8, phase="resolution", steps_per_epoch=300, **kwargs
    ):
        self.resolution = resolution
        self.steps_per_epoch = steps_per_epoch
        self.train_step_counter.assign(0.0)
        self.phase.assign(phase)
        super().fit(*args, steps_per_epoch=steps_per_epoch, **kwargs)

    def save_weights(self, filepath, **kwargs):
        """
        Override base class function to save the weights of the constituent models
        """
        self.generator.save_weights(
            os.path.join(filepath, "g_weights_res_{}.h5".format(self.resolution)),
            **kwargs
        )
        self.discriminator.save_weights(
            os.path.join(filepath, "d_weights_res_{}.h5".format(self.resolution)),
            **kwargs
        )
        self.generator.save(
            os.path.join(filepath, "generator_res_{}".format(self.resolution))
        )


class GANTrainer(tf.keras.Model):
    """Generative Adversarial Network Trainer.

    Trains discriminator and generator alternatively in an adversarial manner for generation of
    brain MRI images.

    Parameters
    ----------
    discriminator : tf.keras.Model, Instantiated using nobrainer.models
    generator : tf.keras.Model, Instantiated using nobrainer.models
    gradient_penalty : boolean, Use gradient penalty on discriminator for smooth training.

    References
    ----------

    Links
    -----
    """

    def __init__(self, discriminator, generator, gradient_penalty=False):
        super(GANTrainer, self).__init__()
        self.discriminator = discriminator
        self.generator = generator
        self.gradient_penalty = gradient_penalty
        self.latent_size = generator.latent_size

    def compile(self, d_optimizer, g_optimizer, g_loss_fn, d_loss_fn):
        super(GANTrainer, self).compile()
        self.d_optimizer = d_optimizer
        self.g_optimizer = g_optimizer

        self.g_loss_fn = compile_utils.LossesContainer(g_loss_fn)
        self.d_loss_fn = compile_utils.LossesContainer(d_loss_fn)

        if self.gradient_penalty:
            self.gradient_penalty_fn = compile_utils.LossesContainer(gradient_penalty)

    def train_step(self, reals):
        if isinstance(reals, tuple):
            reals = reals[0]

        # get batch size dynamically
        batch_size = tf.shape(reals)[0]

        # normalize the real images using minmax to [-1, 1]
        reals = _adjust_dynamic_range(reals, [0.0, 255.0], [-1.0, 1.0])

        # train discriminator
        latents = tf.random.normal((batch_size, self.latent_size))
        fake_labels = tf.ones((batch_size, 1)) * -1
        real_labels = tf.ones((batch_size, 1))

        with tf.GradientTape() as tape:
            fakes = self.generator(latents)
            fakes_pred, labels_pred_fake = self.discriminator(fakes)
            reals_pred, labels_pred_real = self.discriminator(reals)

            fake_loss = self.d_loss_fn(fake_labels, fakes_pred)
            real_loss = self.d_loss_fn(real_labels, reals_pred)
            d_loss = 0.5 * (fake_loss + real_loss)

            # calculate and add the gradient penalty loss using average samples for discriminator
            if self.gradient_penalty:
                weight_shape = (tf.shape(reals)[0],) + (
                    1,
                    1,
                    1,
                    1,
                )  # broadcasting to right shape
                weight = tf.random.uniform(weight_shape, minval=0, maxval=1)
                average_samples = (weight * reals) + ((1 - weight) * fakes)
                average_pred = self.discriminator(average_samples)
                gradients = tf.gradients(average_pred, average_samples)[0]
                gp_loss = self.gradient_penalty_fn(gradients, reals_pred)
                d_loss += gp_loss

        d_gradients = tape.gradient(d_loss, self.discriminator.trainable_variables)
        self.d_optimizer.apply_gradients(
            zip(d_gradients, self.discriminator.trainable_variables)
        )

        # train generator
        misleading_labels = tf.ones((batch_size, 1))

        latents = tf.random.normal((batch_size, self.latent_size))
        with tf.GradientTape() as tape:
            fakes = self.generator(latents)
            fakes_pred, labels_pred = self.discriminator(fakes)

            g_loss = self.g_loss_fn(misleading_labels, fakes_pred)

        g_gradients = tape.gradient(g_loss, self.generator.trainable_variables)
        self.g_optimizer.apply_gradients(
            zip(g_gradients, self.generator.trainable_variables)
        )

        return {"d_loss": d_loss, "g_loss": g_loss}

    def save_weights(self, filepath, **kwargs):
        """
        Override base class function to save the weights of the constituent models
        """
        self.generator.save_weights(
            os.path.join(filepath, "g_weights_res_{}.h5".format(self.resolution)),
            **kwargs
        )
        self.discriminator.save_weights(
            os.path.join(filepath, "d_weights_res_{}.h5".format(self.resolution)),
            **kwargs
        )


class BrainSiamese(tf.keras.Model):
    def __init__(self, encoder, predictor):
        super(BrainSiamese, self).__init__()
        self.encoder = encoder
        self.predictor = predictor
        self.loss_tracker = tf.keras.metrics.Mean(name="loss")

    @property
    def metrics(self):
        return [self.loss_tracker]

    def train_step(self, data):
        data_one, data_two = data  # unpacking the data

        # Forward pass through the encoder and predictor.
        with tf.GradientTape() as tape:
            proj_1, proj_2 = self.encoder(data_one), self.encoder(data_two)
            pred_1, pred_2 = self.predictor(proj_1), self.predictor(proj_2)
            loss = compute_loss(pred_1, proj_2) / 2 + compute_loss(pred_2, proj_1) / 2

        # Compute gradients and update the parameters.
        learnable_params = (
            self.encoder.trainable_variables + self.predictor.trainable_variables
        )
        gradients = tape.gradient(loss, learnable_params)
        self.optimizer.apply_gradients(zip(gradients, learnable_params))

        # Monitor loss.
        self.loss_tracker.update_state(loss)
        return {"loss": self.loss_tracker.result()}


def compute_loss(pred, proj):
    proj = tf.stop_gradient(proj)
    pred = tf.math.l2_normalize(pred, axis=1)
    proj = tf.math.l2_normalize(proj, axis=1)

    # Negative cosine similarity loss
    return -tf.reduce_mean(tf.reduce_sum((pred * proj), axis=1))



"""Tests for `nobrainer.prediction`."""

import nibabel as nib
import numpy as np
from numpy.testing import assert_array_equal
import pytest
import tensorflow as tf

from .. import prediction
from ..models.bayesian import variational_meshnet
from ..models.meshnet import meshnet


def test_predict(tmp_path):

    x = np.ones((4, 4, 4))
    img = nib.Nifti1Image(x, affine=np.eye(4))
    path = str(tmp_path / "features.nii.gz")
    img.to_filename(path)

    x2 = x * -50
    img2 = nib.Nifti1Image(x2, affine=np.eye(4))
    path2 = str(tmp_path / "features2.nii.gz")
    img2.to_filename(path2)

    model = meshnet(1, (*x.shape, 1), receptive_field=37)

    # From array.
    y_ = prediction.predict_from_array(x, model=model, block_shape=None)
    y_blocks = prediction.predict_from_array(x, model=model, block_shape=x.shape)
    y_other = prediction.predict(x, model=model, block_shape=None)
    assert isinstance(y_, np.ndarray)
    assert_array_equal(y_.shape, *x.shape)
    assert_array_equal(y_, y_other)
    assert_array_equal(y_, y_blocks)

    # From image.
    y_img = prediction.predict_from_img(img, model=model, block_shape=None)
    y_img_other = prediction.predict(img, model=model, block_shape=None)
    assert isinstance(y_img, nib.spatialimages.SpatialImage)
    assert_array_equal(y_img.shape, x.shape)
    assert_array_equal(y_img.get_fdata(caching="unchanged"), y_)
    assert_array_equal(
        y_img.get_fdata(caching="unchanged"), y_img_other.get_fdata(caching="unchanged")
    )

    # From filepath
    y_img2 = prediction.predict_from_filepath(path, model=model, block_shape=None)
    y_img2_other = prediction.predict(path, model=model, block_shape=None)
    assert isinstance(y_img, nib.spatialimages.SpatialImage)
    assert_array_equal(y_img.shape, x.shape)
    assert_array_equal(y_img.get_fdata(caching="unchanged"), y_)
    assert_array_equal(
        y_img2.get_fdata(caching="unchanged"),
        y_img2_other.get_fdata(caching="unchanged"),
    )

    # From filepaths
    gen = prediction.predict_from_filepaths(
        [path, path2], model=model, block_shape=None
    )
    y_img3 = next(gen)
    y_img4 = next(gen)
    gen_other = prediction.predict([path, path2], model=model, block_shape=None)
    y_img3_other = next(gen_other)
    y_img4_other = next(gen_other)

    assert_array_equal(
        y_img2.get_fdata(caching="unchanged"), y_img3.get_fdata(caching="unchanged")
    )
    assert_array_equal(
        y_img3.get_fdata(caching="unchanged"),
        y_img3_other.get_fdata(caching="unchanged"),
    )
    assert_array_equal(
        y_img4.get_fdata(caching="unchanged"),
        y_img4_other.get_fdata(caching="unchanged"),
    )
    assert_array_equal(y_img3.shape, x.shape)


def test_variational_predict(tmp_path):
    x = np.ones((4, 4, 4))
    img = nib.Nifti1Image(x, affine=np.eye(4))
    path = str(tmp_path / "features.nii.gz")
    img.to_filename(path)

    x2 = x * -50
    img2 = nib.Nifti1Image(x2, affine=np.eye(4))
    path2 = str(tmp_path / "features2.nii.gz")
    img2.to_filename(path2)

    model = variational_meshnet(1, (*x.shape, 1), receptive_field=37)

    # From array.
    mean, var, entropy = prediction.predict_from_array(
        x,
        model=model,
        block_shape=None,
        n_samples=2,
        return_variance=True,
        return_entropy=True,
    )
    # y_blocks = prediction.predict_from_array(x, model=model, block_shape=x.shape)
    # y_other = prediction.predict(x, model=model, block_shape=None)
    assert isinstance(mean, np.ndarray)
    assert isinstance(var, np.ndarray)
    assert isinstance(entropy, np.ndarray)
    assert_array_equal(mean.shape, *x.shape)
    assert_array_equal(var.shape, *x.shape)
    assert_array_equal(entropy.shape, *x.shape)
    # assert_array_equal(y_, y_other)
    # assert_array_equal(y_, y_blocks)


def test_get_model(tmp_path):
    model = meshnet(3, (10, 10, 10, 1), receptive_field=37)
    path = str(tmp_path / "model.h5")
    model.save(path)
    assert isinstance(prediction._get_model(path), tf.keras.Model)
    assert model is prediction._get_model(model)
    with pytest.raises(ValueError):
        prediction._get_model("not a model")


@pytest.mark.xfail
def test_transform_and_predict(tmp_path):
    assert False



import itertools

import numpy as np
from numpy.testing import assert_array_equal
import pytest
import tensorflow as tf

from .. import transform


@pytest.mark.parametrize("volume_shape", [(64, 64, 64), (64, 64, 64, 3)])
def test_get_affine_smoke(volume_shape):
    affine = transform.get_affine(volume_shape)

    assert_array_equal(affine, np.eye(4))


def test_get_affine_errors():
    with pytest.raises(ValueError):
        transform.get_affine(volume_shape=(64, 64))

    with pytest.raises(ValueError):
        transform.get_affine(volume_shape=(64, 64, 64), rotation=[0, 0])

    with pytest.raises(ValueError):
        transform.get_affine(volume_shape=(64, 64, 64), translation=[0, 0])


@pytest.mark.parametrize("volume_shape", [(2, 2, 2), (2, 2, 2, 3)])
def test_get_coordinates(volume_shape):
    coords = transform._get_coordinates(volume_shape=volume_shape)
    coords_ref = [
        list(element) for element in list(itertools.product([0, 1], repeat=3))
    ]
    assert_array_equal(coords, coords_ref)


def test_get_coordinates_errors():
    with pytest.raises(ValueError):
        transform._get_coordinates(volume_shape=(64, 64))


@pytest.mark.parametrize("volume_shape", [(8, 8, 8), (8, 8, 8, 3)])
def test_trilinear_interpolation_smoke(volume_shape):
    volume = np.arange(np.prod(volume_shape)).reshape(volume_shape)
    coords = transform._get_coordinates(volume_shape=volume_shape)
    x = transform._trilinear_interpolation(volume=volume, coords=coords)
    assert_array_equal(x, volume)


@pytest.mark.parametrize("volume_shape", [(8, 8, 8), (8, 8, 8, 3)])
def test_get_voxels(volume_shape):
    volume = np.arange(np.prod(volume_shape)).reshape(volume_shape)
    coords = transform._get_coordinates(volume_shape=volume_shape)
    voxels = transform._get_voxels(volume=volume, coords=coords)

    if len(volume_shape) == 3:
        assert_array_equal(voxels, np.arange(np.prod(volume_shape)))
    else:
        assert_array_equal(
            voxels,
            np.arange(np.prod(volume_shape)).reshape((np.prod(volume_shape[:3]), -1)),
        )


def test_get_voxels_errors():
    volume = np.zeros((8, 8))
    coords = transform._get_coordinates(volume_shape=(8, 8, 8))
    with pytest.raises(ValueError):
        transform._get_voxels(volume=volume, coords=coords)

    volume = np.zeros((8, 8, 8))
    coords = np.zeros((8, 8, 8))
    with pytest.raises(ValueError):
        transform._get_voxels(volume=volume, coords=coords)

    coords = np.zeros((8, 2))
    with pytest.raises(ValueError):
        transform._get_voxels(volume=volume, coords=coords)


@pytest.mark.parametrize("shape", [(10, 10, 10), (10, 10, 10, 3)])
@pytest.mark.parametrize("scalar_labels", [True, False])
def test_apply_random_transform(shape, scalar_labels):
    x = np.ones(shape).astype(np.float32)
    if scalar_labels:
        transform_func = transform.apply_random_transform_scalar_labels
        y_shape = (1,)
    else:
        transform_func = transform.apply_random_transform
        y_shape = shape

    y_in = np.random.randint(0, 2, size=y_shape).astype(np.float32)
    x, y = transform_func(x, y_in)
    x = x.numpy()
    y = y.numpy()

    # Test that values were not changed in the labels.
    if scalar_labels:
        assert_array_equal(y, y_in)
    else:
        assert_array_equal(np.unique(y), [0, 1])
    assert x.shape == shape
    assert y.shape == y_shape

    with pytest.raises(ValueError):
        inconsistent_shape = tuple([sh + 1 for sh in shape])
        x, y = transform_func(np.ones(shape), np.ones(inconsistent_shape))

    with pytest.raises(ValueError):
        y_shape = (1,) if scalar_labels else (10, 10)
        x, y = transform_func(np.ones((10, 10)), np.ones(y_shape))

    x = np.random.randn(*shape).astype(np.float32)
    y_shape = (1,) if scalar_labels else shape
    y = np.random.randint(0, 2, size=y_shape).astype(np.float32)
    x0, y0 = transform_func(x, y)
    x1, y1 = transform_func(x, y)
    assert not np.array_equal(x, x0)
    assert not np.array_equal(x, x1)
    assert not np.array_equal(x0, x1)

    if scalar_labels:
        assert np.array_equal(y, y0)
        assert np.array_equal(y, y1)
        assert np.array_equal(y0, y1)
    else:
        assert not np.array_equal(y, y0)
        assert not np.array_equal(y, y1)
        assert not np.array_equal(y0, y1)

    # Test that new iterations yield different augmentations.
    x = np.arange(64).reshape(1, 4, 4, 4).astype(np.float32)
    y_shape = (1, 1) if scalar_labels else x.shape
    y = np.random.randint(0, 2, size=y_shape).astype(np.float32)
    dataset = tf.data.Dataset.from_tensor_slices((x, y))
    # sanity check
    x0, y0 = next(iter(dataset))
    x1, y1 = next(iter(dataset))
    assert_array_equal(x[0], x0)
    assert_array_equal(x0, x1)
    assert_array_equal(y[0], y0)
    assert_array_equal(y0, y1)
    # Need to reset the seed, because it is set in other tests.
    tf.random.set_seed(None)
    dataset = dataset.map(transform_func)
    x0, y0 = next(iter(dataset))
    x1, y1 = next(iter(dataset))
    assert not np.array_equal(x0, x1)
    if scalar_labels:
        assert_array_equal(y0, y1)
    else:
        assert not np.array_equal(y0, y1)
        assert_array_equal(np.unique(y0), [0, 1])
        assert_array_equal(np.unique(y1), [0, 1])
    # Naive test that features were interpolated without nearest neighbor.
    assert np.any(x0 % 1)
    assert np.any(x1 % 1)



"""Volumetric affine transformations implemented in TensorFlow."""

import numpy as np
import tensorflow as tf


def warp_features_labels(features, labels, matrix, scalar_label=False):
    """Warp features and labels tensors according to affine matrix.

    Trilinear interpolation is used for features, and nearest neighbor
    interpolation is used for labels.

    Parameters
    ----------
    features: Rank 3 tensor, volumetric feature data.
    labels: Rank 3 tensor or N
    matrix: Tensor with shape `(4, 4)`, affine matrix.

    Returns
    -------
    Tuple of warped features, warped labels.
    """
    features = tf.convert_to_tensor(features)
    labels = tf.convert_to_tensor(labels)

    warped_coords = _warp_coords(matrix=matrix, volume_shape=features.shape)
    features = _trilinear_interpolation(volume=features, coords=warped_coords)
    if not scalar_label:
        labels = _nearest_neighbor_interpolation(volume=labels, coords=warped_coords)
    return (features, labels)


def warp(volume, matrix, order=1):
    """Warp volume tensor according to affine matrix.

    Parameters
    ----------
    volume: Rank 3 tensor, volume data.
    matrix: Tensor with shape `(4, 4)`, affine matrix.
    order: {0, 1}, interpolation order. 0 is nearest neighbor, and 1 is
        trilinear.

    Returns
    -------
    Tensor of warped volume data.
    """
    volume = tf.convert_to_tensor(volume)

    warped_coords = _warp_coords(matrix=matrix, volume_shape=volume.shape)
    if order == 0:
        out = _nearest_neighbor_interpolation(volume=volume, coords=warped_coords)
    elif order == 1:
        out = _trilinear_interpolation(volume=volume, coords=warped_coords)
    else:
        raise ValueError("unknown 'order'. Valid values are 0 and 1.")
    return out


def _warp_coords(matrix, volume_shape):
    """Build the coordinates for a affine transform on volumetric data.

    Parameters
    ----------
    matrix: tensor with shape (4, 4), affine matrix.
    volume_shape: tuple of length 3, shape of output volume.

    Returns
    -------
    TODO check this.
    Tensor of coordinates with shape (*volume_shape, 3).
    """
    coords = _get_coordinates(volume_shape=volume_shape)
    # Append ones to play nicely with 4x4 affine.
    coords_homogeneous = tf.concat(
        [coords, tf.ones((coords.shape[0], 1), dtype=coords.dtype)], axis=1
    )
    return (coords_homogeneous @ tf.transpose(matrix))[..., :3]


def get_affine(volume_shape, rotation=[0, 0, 0], translation=[0, 0, 0]):
    """Return 4x4 affine, which encodes rotation and translation of 3D tensors.

    Parameters
    ----------
    rotation: iterable of three numbers, the yaw, pitch, and roll,
        respectively, in radians.
    translation: iterable of three numbers, the number of voxels to translate
        in the x, y, and z directions.

    Returns
    -------
    Tensor with shape `(4, 4)` and dtype float32.
    """
    volume_shape = tf.cast(volume_shape, tf.float32)
    rotation = tf.cast(rotation, tf.float32)
    translation = tf.cast(translation, tf.float32)
    if volume_shape.shape[0] < 3:
        raise ValueError("`volume_shape` must have at least three values")
    if rotation.shape[0] != 3:
        raise ValueError("`rotation` must have three values")
    if translation.shape[0] != 3:
        raise ValueError("`translation` must have three values")

    # ROTATION
    # yaw
    rx = tf.convert_to_tensor(
        [
            [1, 0, 0, 0],
            [0, tf.math.cos(rotation[0]), -tf.math.sin(rotation[0]), 0],
            [0, tf.math.sin(rotation[0]), tf.math.cos(rotation[0]), 0],
            [0, 0, 0, 1],
        ],
        dtype=tf.float32,
    )

    # pitch
    ry = tf.convert_to_tensor(
        [
            [tf.math.cos(rotation[1]), 0, tf.math.sin(rotation[1]), 0],
            [0, 1, 0, 0],
            [-tf.math.sin(rotation[1]), 0, tf.math.cos(rotation[1]), 0],
            [0, 0, 0, 1],
        ],
        dtype=tf.float32,
    )

    # roll
    rz = tf.convert_to_tensor(
        [
            [tf.math.cos(rotation[2]), -tf.math.sin(rotation[2]), 0, 0],
            [tf.math.sin(rotation[2]), tf.math.cos(rotation[2]), 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1],
        ],
        dtype=tf.float32,
    )

    # Rotation around origin.
    transform = rz @ ry @ rx

    center = tf.convert_to_tensor(volume_shape[:3] / 2 - 0.5, dtype=tf.float32)
    neg_center = tf.math.negative(center)
    center_to_origin = tf.convert_to_tensor(
        [
            [1, 0, 0, neg_center[0]],
            [0, 1, 0, neg_center[1]],
            [0, 0, 1, neg_center[2]],
            [0, 0, 0, 1],
        ],
        dtype=tf.float32,
    )

    origin_to_center = tf.convert_to_tensor(
        [
            [1, 0, 0, center[0]],
            [0, 1, 0, center[1]],
            [0, 0, 1, center[2]],
            [0, 0, 0, 1],
        ],
        dtype=tf.float32,
    )

    # Rotation around center of volume.
    transform = origin_to_center @ transform @ center_to_origin

    # TRANSLATION
    translation = tf.convert_to_tensor(
        [
            [1, 0, 0, translation[0]],
            [0, 1, 0, translation[1]],
            [0, 0, 1, translation[2]],
            [0, 0, 0, 1],
        ],
        dtype=tf.float32,
    )

    transform = translation @ transform

    # REFLECTION
    #
    # TODO.
    # See http://web.iitd.ac.in/~hegde/cad/lecture/L6_3dtrans.pdf#page=7
    # and https://en.wikipedia.org/wiki/Transformation_matrix#Reflection_2

    return transform


def _get_coordinates(volume_shape):
    """Get coordinates that represent every voxel in a volume with shape
    `volume_shape`.

    Parameters
    ----------
    volume_shape: tuple of length 3, shape of output volume.

    Returns
    -------
    Tensor of coordinates with shape `(prod(volume_shape), 3)`.
    """
    if len(volume_shape) < 3:
        raise ValueError("shape must have at least 3 items.")
    dtype = tf.float32
    rows, cols, depth = volume_shape[:3]

    out = tf.meshgrid(
        tf.range(rows, dtype=dtype),
        tf.range(cols, dtype=dtype),
        tf.range(depth, dtype=dtype),
        indexing="ij",
    )
    return tf.reshape(tf.stack(out, axis=3), shape=(-1, 3))


def _nearest_neighbor_interpolation(volume, coords):
    """Three-dimensional nearest neighbors interpolation."""
    volume_f = _get_voxels(volume=volume, coords=tf.round(coords))
    return tf.reshape(volume_f, volume.shape)


def _trilinear_interpolation(volume, coords):
    """Trilinear interpolation.

    Implemented according to
    https://en.wikipedia.org/wiki/Trilinear_interpolation#Method
    https://github.com/Ryo-Ito/spatial_transformer_network/blob/2555e846b328e648a456f92d4c80fce2b111599e/warp.py#L137-L222
    """
    volume = tf.cast(volume, tf.float32)
    coords = tf.cast(coords, tf.float32)
    coords_floor = tf.floor(coords)

    shape = tf.shape(volume)
    xlen = shape[0]
    ylen = shape[1]
    zlen = shape[2]

    # Get lattice points. x0 is point below x, and x1 is point above x. Same for y and
    # z.
    x0 = tf.cast(coords_floor[:, 0], tf.int32)
    x1 = x0 + 1
    y0 = tf.cast(coords_floor[:, 1], tf.int32)
    y1 = y0 + 1
    z0 = tf.cast(coords_floor[:, 2], tf.int32)
    z1 = z0 + 1

    # Clip values to the size of the volume array.
    x0 = tf.clip_by_value(x0, 0, xlen - 1)
    x1 = tf.clip_by_value(x1, 0, xlen - 1)
    y0 = tf.clip_by_value(y0, 0, ylen - 1)
    y1 = tf.clip_by_value(y1, 0, ylen - 1)
    z0 = tf.clip_by_value(z0, 0, zlen - 1)
    z1 = tf.clip_by_value(z1, 0, zlen - 1)

    # Get the indices at corners of cube.
    i000 = x0 * ylen * zlen + y0 * zlen + z0
    i001 = x0 * ylen * zlen + y0 * zlen + z1
    i010 = x0 * ylen * zlen + y1 * zlen + z0
    i011 = x0 * ylen * zlen + y1 * zlen + z1
    i100 = x1 * ylen * zlen + y0 * zlen + z0
    i101 = x1 * ylen * zlen + y0 * zlen + z1
    i110 = x1 * ylen * zlen + y1 * zlen + z0
    i111 = x1 * ylen * zlen + y1 * zlen + z1

    # Get volume values at corners of cube.
    if len(volume.shape) == 3:
        volume_flat = tf.reshape(volume, [-1])
    else:
        volume_flat = tf.reshape(volume, [-1, volume.shape[-1]])

    c000 = tf.gather(volume_flat, i000)
    c001 = tf.gather(volume_flat, i001)
    c010 = tf.gather(volume_flat, i010)
    c011 = tf.gather(volume_flat, i011)
    c100 = tf.gather(volume_flat, i100)
    c101 = tf.gather(volume_flat, i101)
    c110 = tf.gather(volume_flat, i110)
    c111 = tf.gather(volume_flat, i111)

    xd = coords[:, 0] - tf.cast(x0, tf.float32)
    yd = coords[:, 1] - tf.cast(y0, tf.float32)
    zd = coords[:, 2] - tf.cast(z0, tf.float32)

    if len(volume.shape) == 4:
        # Add a channels axis for proper broadcasting
        xd = xd[:, tf.newaxis]
        yd = yd[:, tf.newaxis]
        zd = zd[:, tf.newaxis]

    # Interpolate along x-axis.
    c00 = c000 * (1 - xd) + c100 * xd
    c01 = c001 * (1 - xd) + c101 * xd
    c10 = c010 * (1 - xd) + c110 * xd
    c11 = c011 * (1 - xd) + c111 * xd

    # Interpolate along y-axis.
    c0 = c00 * (1 - yd) + c10 * yd
    c1 = c01 * (1 - yd) + c11 * yd

    # Interpolate along z-axis.
    c = c0 * (1 - zd) + c1 * zd

    return tf.reshape(c, volume.shape)


def _get_voxels(volume, coords):
    """Get voxels from volume at points. These voxels are in a flat tensor."""
    x = tf.cast(volume, tf.float32)
    coords = tf.cast(coords, tf.float32)

    if len(x.shape) < 3:
        raise ValueError("`volume` must be at least rank 3")
    if len(coords.shape) != 2 or coords.shape[1] != 3:
        raise ValueError("`coords` must have shape `(N, 3)`.")

    rows, cols, depth, *n_channels = x.shape

    # Points in flattened array representation.
    fcoords = coords[:, 0] * cols * depth + coords[:, 1] * depth + coords[:, 2]

    # Some computed finds are out of range of the image's flattened size.
    # Zero those so we don't get errors. These points in the volume are filled later.
    fcoords_size = tf.size(fcoords, out_type=fcoords.dtype)
    fcoords = tf.clip_by_value(fcoords, 0, fcoords_size - 1)
    xflat = tf.squeeze(tf.reshape(x, [tf.math.reduce_prod(x.shape[:3]), -1]))

    # Reorder image data to transformed space.
    xflat = tf.gather(params=xflat, indices=tf.cast(fcoords, tf.int32))

    # Zero image data that was out of frame.
    outofframe = (
        tf.reduce_any(coords < 0, -1)
        | (coords[:, 0] > rows)
        | (coords[:, 1] > cols)
        | (coords[:, 2] > depth)
    )

    if n_channels:
        outofframe = tf.stack([outofframe for _ in range(n_channels[0])], axis=-1)

    xflat = tf.multiply(xflat, tf.cast(tf.logical_not(outofframe), xflat.dtype))

    return xflat


def apply_random_transform(features, labels):
    """Apply a random rigid transformation to `features` and `labels`.

    The same transformation is applied to features and labels. Features are
    interpolated trilinearly, and labels are interpolated with nearest
    neighbors.
    """
    if len(features.shape) < 3 or len(labels.shape) < 3:
        raise ValueError("features and labels must be at least rank 3")
    if features.shape != labels.shape:
        raise ValueError("shape of features and labels must be the same.")
    # Rotate -180 degrees to 180 degrees in three dimensions.
    rotation = tf.random.uniform(
        shape=[3], minval=-np.pi, maxval=np.pi, dtype=tf.float32
    )

    # Translate at most 5% in any direction, so there's less chance of
    # important data going out of view.
    maxval = 0.05 * features.shape[0]
    translation = tf.random.uniform(
        shape=[3], minval=-maxval, maxval=maxval, dtype=tf.float32
    )

    volume_shape = np.asarray(features.shape)
    matrix = get_affine(
        volume_shape=volume_shape, rotation=rotation, translation=translation
    )
    return warp_features_labels(features=features, labels=labels, matrix=matrix)


def apply_random_transform_scalar_labels(features, labels):
    """Apply a random rigid transformation to `features`.

    Features are interpolated trilinearly, and labels are unchanged because they are
    scalars.
    """
    if len(features.shape) < 3:
        raise ValueError("features must be at least rank 3")
    if len(labels.shape) != 1:
        raise ValueError("labels must be rank 1")
    # Rotate -180 degrees to 180 degrees in three dimensions.
    rotation = tf.random.uniform(
        shape=[3], minval=-np.pi, maxval=np.pi, dtype=tf.float32
    )

    # Translate at most 5% in any direction, so there's less chance of
    # important data going out of view.
    maxval = 0.05 * features.shape[0]
    translation = tf.random.uniform(
        shape=[3], minval=-maxval, maxval=maxval, dtype=tf.float32
    )

    volume_shape = np.asarray(features.shape)
    matrix = get_affine(
        volume_shape=volume_shape, rotation=rotation, translation=translation
    )
    return warp_features_labels(
        features=features, labels=labels, matrix=matrix, scalar_label=True
    )



from pathlib import Path

import numpy as np
from numpy.testing import assert_allclose

from .. import utils as nbutils
from ..io import read_csv


def test_get_data():
    csv_path = nbutils.get_data()
    assert Path(csv_path).is_file()

    files = read_csv(csv_path)
    assert len(files) == 10
    assert all(len(r) == 2 for r in files)
    for x, y in files:
        assert Path(x).is_file()
        assert Path(y).is_file()


def test_streaming_stats():
    # TODO: add entropy
    ss = nbutils.StreamingStats()
    xs = np.random.random_sample((100))
    for x in xs:
        ss.update(x)
    assert_allclose(xs.mean(), ss.mean())
    assert_allclose(xs.std(), ss.std())
    assert_allclose(xs.var(), ss.var())

    ss = nbutils.StreamingStats()
    xs = np.random.random_sample((10, 5, 5, 5))
    for x in xs:
        ss.update(x)
    assert_allclose(xs.mean(0), ss.mean())
    assert_allclose(xs.std(0), ss.std())
    assert_allclose(xs.var(0), ss.var())



"""Implementations of Bayesian neural networks."""

import tensorflow as tf
import tensorflow_probability as tfp

from ..layers.dropout import BernoulliDropout, ConcreteDropout

# TODO: add WeightNorm wrapper from tensorflow-probability once next
# version of tfp is released.

tfk = tf.keras
tfkl = tfk.layers
tfpl = tfp.layers


def variational_meshnet(
    n_classes,
    input_shape,
    receptive_field=67,
    filters=71,
    is_monte_carlo=False,
    dropout=None,
    activation=tf.nn.relu,
    batch_size=None,
    name="variational_meshnet",
):
    """Instantiate variational MeshNet model.

    Please see https://arxiv.org/abs/1805.10863 for more information.

    Parameters
    ----------
    n_classes: int, number of classes to classify. For binary applications, use
        a value of 1.
    input_shape: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    receptive_field: {37, 67, 129}, the receptive field of the model. According
        to the MeshNet manuscript, the receptive field should be similar to your
        input shape. The actual receptive field is the cube of the value provided.
    filters: int, number of filters per volumetric convolution. The original
        MeshNet manuscript uses 21 filters for a binary segmentation task
        (i.e., brain extraction) and 71 filters for a multi-class segmentation task.
    activation: str or optimizer object, the non-linearity to use.
    is_monte_carlo: bool, if true, apply variational approximation in the variational
        convolution layers. If false, will use the most likely weight (i.e.,
        the mean of the weight gaussian) instead of sampling a weight from the
        gaussian distribution.
    dropout: string, type of dropout layer.
    batch_size: int, number of samples in each batch. This must be set when
        training on TPUs.
    name: str, name to give to the resulting model object.

    Returns
    -------
    Model object.

    Raises
    ------
    ValueError if receptive field is not an allowable value.
    """

    if receptive_field not in {37, 67, 129}:
        raise ValueError("unknown receptive field. Legal values are 37, 67, and 129.")

    def one_layer(x, layer_num, dilation_rate=(1, 1, 1)):
        # TODO: implement correct behavior for is_mc.
        x = tfpl.Convolution3DReparameterization(
            filters,
            kernel_size=3,
            padding="same",
            dilation_rate=dilation_rate,
            activation=activation,
            name="layer{}/vwnconv3d".format(layer_num),
        )(x)
        if dropout is None:
            pass
        elif dropout == "bernoulli":
            x = BernoulliDropout(
                rate=0.5,
                is_monte_carlo=is_monte_carlo,
                scale_during_training=False,
                name="layer{}/bernoulli_dropout".format(layer_num),
            )(x)
        elif dropout == "concrete":
            x = ConcreteDropout(
                is_monte_carlo=is_monte_carlo,
                temperature=0.02,
                use_expectation=is_monte_carlo,
                name="layer{}/concrete_dropout".format(layer_num),
            )(x)
        else:
            raise ValueError("unknown dropout layer, {}".format(dropout))
        x = tfkl.Activation(activation, name="layer{}/activation".format(layer_num))(x)
        return x

    inputs = tfkl.Input(shape=input_shape, batch_size=batch_size, name="inputs")

    if receptive_field == 37:
        x = one_layer(inputs, 1)
        x = one_layer(x, 2)
        x = one_layer(x, 3)
        x = one_layer(x, 4, dilation_rate=(2, 2, 2))
        x = one_layer(x, 5, dilation_rate=(4, 4, 4))
        x = one_layer(x, 6, dilation_rate=(8, 8, 8))
        x = one_layer(x, 7)
    elif receptive_field == 67:
        x = one_layer(inputs, 1)
        x = one_layer(x, 2)
        x = one_layer(x, 3, dilation_rate=(2, 2, 2))
        x = one_layer(x, 4, dilation_rate=(4, 4, 4))
        x = one_layer(x, 5, dilation_rate=(8, 8, 8))
        x = one_layer(x, 6, dilation_rate=(16, 16, 16))
        x = one_layer(x, 7)
    elif receptive_field == 129:
        x = one_layer(inputs, 1)
        x = one_layer(x, 2, dilation_rate=(2, 2, 2))
        x = one_layer(x, 3, dilation_rate=(4, 4, 4))
        x = one_layer(x, 4, dilation_rate=(8, 8, 8))
        x = one_layer(x, 5, dilation_rate=(16, 16, 16))
        x = one_layer(x, 6, dilation_rate=(32, 32, 32))
        x = one_layer(x, 7)

    # TODO: implement is_mc behavior.
    x = tfpl.Convolution3DReparameterization(
        filters=n_classes,
        kernel_size=1,
        padding="same",
        name="classification/vwnconv3d",
    )(x)

    final_activation = "sigmoid" if n_classes == 1 else "softmax"
    x = tfkl.Activation(final_activation, name="classification/activation")(x)

    return tf.keras.Model(inputs=inputs, outputs=x, name=name)



"""Methods to predict using trained models."""

import math
from pathlib import Path

import nibabel as nib
import numpy as np
import tensorflow as tf

from .transform import get_affine, warp
from .utils import StreamingStats
from .volume import from_blocks, from_blocks_numpy, standardize_numpy, to_blocks_numpy


def predict(
    inputs,
    model,
    block_shape,
    batch_size=1,
    normalizer=None,
    n_samples=1,
    return_variance=False,
    return_entropy=False,
):
    """Return predictions from `inputs`.

    This is a general prediction method that can accept various types of
    `inputs`.

    Parameters
    ----------
    inputs: 3D array or Nibabel image or filepath or list of filepaths.
    model: str: path to saved model, either HDF5 or SavedModel.
    block_shape: tuple of length 3, shape of sub-volumes on which to
        predict.
    batch_size: int, number of sub-volumes per batch for predictions.
    normalizer: callable, function that accepts an ndarray and returns an
        ndarray. Called before separating volume into blocks.
    n_samples: The number of sampling. If set as 1, it will just return the
        single prediction value. The default value is 1
    return_variance: Boolean. If set True, it returns the running population
        variance along with mean. Note, if the n_samples is smaller or equal to 1,
        the variance will not be returned; instead it will return None
    return_entropy: Boolean. If set True, it returns the running entropy.
        along with mean.

    Returns
    -------
    If `inputs` is a:
        - 3D numpy array, return an iterable of maximum 3 elements;
            3D array of mean, variance(optional),and entropy(optional) of prediction.
            if the flags for variance or entropy is set False, it won't be returned at
            all The specific order of the elements are:
            mean, variance(default=None) , entropy(default=None)
            Note, variance is only defined when n_sample  > 1
        - Nibabel image or filepath, return a set of Nibabel images of mean, variance,
            entropy of predictions or just the pure arrays of them,
            if return_array_from_images is True.
        - list of filepaths, return generator that yields one set of Nibabel images
            or arrays(if return_array_from_images is set True) of means, variance, and
            entropy predictions per iteration.
    """
    if n_samples < 1:
        raise Exception("n_samples cannot be lower than 1.")

    model = _get_model(model)

    if isinstance(inputs, np.ndarray):
        out = predict_from_array(
            inputs=inputs,
            model=model,
            block_shape=block_shape,
            batch_size=batch_size,
            normalizer=normalizer,
            n_samples=n_samples,
            return_variance=return_variance,
            return_entropy=return_entropy,
        )
    elif isinstance(inputs, nib.spatialimages.SpatialImage):
        out = predict_from_img(
            img=inputs,
            model=model,
            block_shape=block_shape,
            batch_size=batch_size,
            normalizer=normalizer,
            n_samples=n_samples,
            return_variance=return_variance,
            return_entropy=return_entropy,
        )
    elif isinstance(inputs, str):
        out = predict_from_filepath(
            filepath=inputs,
            model=model,
            block_shape=block_shape,
            batch_size=batch_size,
            normalizer=normalizer,
            n_samples=n_samples,
            return_variance=return_variance,
            return_entropy=return_entropy,
        )
    elif isinstance(inputs, (list, tuple)):
        out = predict_from_filepaths(
            filepaths=inputs,
            model=model,
            block_shape=block_shape,
            batch_size=batch_size,
            normalizer=normalizer,
            n_samples=n_samples,
            return_variance=return_variance,
            return_entropy=return_entropy,
        )
    else:
        raise TypeError("Input to predict is not a valid type")
    return out


def predict_from_array(
    inputs,
    model,
    block_shape,
    batch_size=1,
    normalizer=None,
    n_samples=1,
    return_variance=False,
    return_entropy=False,
):
    """Return a prediction given a filepath and an ndarray of features.

    Parameters
    ----------
    inputs: ndarray, array of features.
    model: `tf.keras.Model`, trained model.
    block_shape: tuple of length 3, shape of sub-volumes on which to
        predict.
    batch_size: int, number of sub-volumes per batch for predictions.
    normalizer: callable, function that accepts an ndarray and returns an
        ndarray. Called before separating volume into blocks.
    n_samples: The number of sampling. If set as 1, it will just return the
        single prediction value. The default value is 1
    return_variance: Boolean. If set True, it returns the running population
        variance along with mean. Note, if the n_samples is smaller or equal to 1,
        the variance will not be returned; instead it will return None
    return_entropy: Boolean. If set True, it returns the running entropy.
        along with mean.

    Returns
    -------
    ndarray of predictions.
    """
    if normalizer:
        features = normalizer(inputs)
    else:
        features = inputs
    if block_shape is not None:
        features = to_blocks_numpy(features, block_shape=block_shape)
    else:
        features = features[None]  # Add batch dimension.

    # Add a dimension for single channel.
    features = features[..., None]

    # Predict per block to reduce memory consumption.
    n_blocks = features.shape[0]
    n_batches = math.ceil(n_blocks / batch_size)

    if not return_variance and not return_entropy and n_samples == 1:
        # TODO: has better performance but output should change to numpy array
        # outputs = model(features).numpy()
        outputs = model.predict(features, batch_size=1, verbose=0)
        if outputs.shape[-1] == 1:
            # Binarize according to threshold.
            outputs = outputs > 0.3
            outputs = outputs.squeeze(-1)
            # Nibabel doesn't like saving boolean arrays as Nifti.
            outputs = outputs.astype(np.uint8)
        else:
            # Hard classes for multi-class segmentation.
            outputs = np.argmax(outputs, -1)
        outputs = from_blocks_numpy(outputs, output_shape=inputs.shape)
        return outputs
    else:
        # Bayesian prediction based on sampling from distributions
        means = np.zeros_like(features.squeeze(-1))
        variances = np.zeros_like(features.squeeze(-1))
        entropies = np.zeros_like(features.squeeze(-1))
        progbar = tf.keras.utils.Progbar(n_batches)
        progbar.update(0)
        for j in range(0, n_blocks, batch_size):

            this_x = features[j : j + batch_size]
            s = StreamingStats()
            for n in range(n_samples):
                # TODO: has better performance but output should change to numpy array
                # new_prediction = model(this_x).numpy()
                new_prediction = model.predict(this_x, batch_size=1, verbose=0)
                s.update(new_prediction)

            means[j : j + batch_size] = np.argmax(s.mean(), axis=-1)  # max mean
            variances[j : j + batch_size] = np.sum(s.var(), axis=-1)
            entropies[j : j + batch_size] = np.sum(s.entropy(), axis=-1)  # entropy
            progbar.add(1)

        total_means = from_blocks_numpy(means, output_shape=inputs.shape)
        total_variance = from_blocks_numpy(variances, output_shape=inputs.shape)
        total_entropy = from_blocks_numpy(entropies, output_shape=inputs.shape)

    include_variance = (n_samples > 1) and (return_variance)
    if include_variance:
        if return_entropy:
            return total_means, total_variance, total_entropy
        else:
            return total_means, total_variance
    else:
        if return_entropy:
            return total_means, total_entropy
        else:
            return total_means


def predict_from_img(
    img,
    model,
    block_shape,
    batch_size=1,
    normalizer=None,
    n_samples=1,
    return_variance=False,
    return_entropy=False,
):
    """Return a prediction given a Nibabel image instance and a predictor.

    Parameters
    ----------
    img: nibabel image, image on which to predict.
    model: `tf.keras.Model`, trained model.
    block_shape: tuple of length 3, shape of sub-volumes on which to
        predict.
    batch_size: int, number of sub-volumes per batch for predictions.
    normalizer: callable, function that accepts an ndarray and returns an
        ndarray. Called before separating volume into blocks.
    n_samples: The number of sampling. If set as 1, it will just return the
        single prediction value. The default value is 1
    return_variance: Boolean. If set True, it returns the running population
        variance along with mean. Note, if the n_samples is smaller or equal to 1,
        the variance will not be returned; instead it will return None
    return_entropy: Boolean. If set True, it returns the running entropy.
        along with mean.

    Returns
    -------
    `nibabel.spatialimages.SpatialImage` or arrays of prediction of mean,
    variance(optional) or entropy (optional).
    """
    if not isinstance(img, nib.spatialimages.SpatialImage):
        raise ValueError("image is not a nibabel image type")
    inputs = np.asarray(img.dataobj)
    img.uncache()
    inputs = inputs.astype(np.float32)

    y = predict_from_array(
        inputs=inputs,
        model=model,
        block_shape=block_shape,
        batch_size=batch_size,
        normalizer=normalizer,
        n_samples=n_samples,
        return_variance=return_variance,
        return_entropy=return_entropy,
    )

    if isinstance(y, np.ndarray):
        return nib.spatialimages.SpatialImage(
            dataobj=y, affine=img.affine, header=img.header, extra=img.extra
        )
    else:
        if len(y) == 2:
            return (
                nib.spatialimages.SpatialImage(
                    dataobj=y[0], affine=img.affine, header=img.header, extra=img.extra
                ),
                nib.spatialimages.SpatialImage(
                    dataobj=y[1], affine=img.affine, header=img.header, extra=img.extra
                ),
            )
        elif len(y) == 3:
            return (
                nib.spatialimages.SpatialImage(
                    dataobj=y[0], affine=img.affine, header=img.header, extra=img.extra
                ),
                nib.spatialimages.SpatialImage(
                    dataobj=y[1], affine=img.affine, header=img.header, extra=img.extra
                ),
                nib.spatialimages.SpatialImage(
                    dataobj=y[2], affine=img.affine, header=img.header, extra=img.extra
                ),
            )


def predict_from_filepath(
    filepath,
    model,
    block_shape,
    batch_size=1,
    normalizer=None,
    n_samples=1,
    return_variance=False,
    return_entropy=False,
):
    """Predict on a volume given a filepath and a trained model.

    Parameters
    ----------
    filepath: path-like, path to existing neuroimaging volume on which
        to predict.
    model: `tf.keras.Model`, trained model.
    block_shape: tuple of length 3, shape of sub-volumes on which to
        predict.
    batch_size: int, number of sub-volumes per batch for predictions.
    normalizer: callable, function that accepts an ndarray and returns an
        ndarray. Called before separating volume into blocks.
    n_samples: The number of sampling. If set as 1, it will just return the
        single prediction value. The default value is 1
    return_variance: Boolean. If set True, it returns the running population
        variance along with mean. Note, if the n_samples is smaller or equal to 1,
        the variance will not be returned; instead it will return None
    return_entropy: Boolean. If set True, it returns the running entropy.
        along with mean.

    Returns
    -------
    `nibabel.spatialimages.SpatialImage` or arrays of predictions of
        mean, variance(optional), and entropy (optional).
    """
    if not Path(filepath).is_file():
        raise FileNotFoundError("could not find file {}".format(filepath))
    img = nib.load(filepath)
    return predict_from_img(
        img=img,
        model=model,
        block_shape=block_shape,
        batch_size=batch_size,
        normalizer=normalizer,
        n_samples=n_samples,
        return_variance=return_variance,
        return_entropy=return_entropy,
    )


def predict_from_filepaths(
    filepaths,
    model,
    block_shape,
    batch_size=1,
    normalizer=None,
    n_samples=1,
    return_variance=False,
    return_entropy=False,
):
    """Yield a model's predictions on a list of filepaths.

    Parameters
    ----------
    filepaths: list, volume filepaths on which to predict.
    model: `tf.keras.Model`, trained model.
    block_shape: tuple of length 3, shape of sub-volumes on which to
        predict.
    batch_size: int, number of sub-volumes per batch for predictions.
    normalizer: callable, function that accepts an ndarray and returns an
        ndarray. Called before separating volume into blocks.
    n_samples: The number of sampling. If set as 1, it will just return the
        single prediction value. The default value is 1
    return_variance: Boolean. If set True, it returns the running population
        variance along with mean. Note, if the n_samples is smaller or equal to 1,
        the variance will not be returned; instead it will return None
    return_entropy: Boolean. If set True, it returns the running entropy.
        along with mean.

    Returns
    -------
    Generator object that yields a `nibabel.spatialimages.SpatialImage` or
    arrays of predictions per filepath in list of input filepaths.
    """
    for filepath in filepaths:
        yield predict_from_filepath(
            filepath=filepath,
            model=model,
            block_shape=block_shape,
            batch_size=batch_size,
            normalizer=normalizer,
            n_samples=n_samples,
            return_variance=return_variance,
            return_entropy=return_entropy,
        )


def _get_model(path):
    """Return `tf.keras.Model` object from a filepath.

    Parameters
    ----------
    path: str, path to HDF5 or SavedModel file.

    Returns
    -------
    Instance of `tf.keras.Model`.

    Raises
    ------
    `ValueError` if cannot load model.
    """
    if isinstance(path, tf.keras.Model):
        return path
    try:
        return tf.keras.models.load_model(path, compile=False)
    except OSError:
        # Not an HDF5 file.
        pass

    try:
        path = Path(path)
        if path.suffix == ".json":
            path = path.parent.parent
        return tf.keras.experimental.load_from_saved_model(str(path))
    except Exception:
        pass

    raise ValueError(
        "Failed to load model. Is the model in HDF5 format or SavedModel" " format?"
    )


def _get_predictor(model_path):
    """restores the tf estimator model.
    The output is a tf2 estimator"""
    predictor = tf.saved_model.load(model_path)
    return predictor.signatures["serving_default"]


def _transform_and_predict(
    model, x, block_shape, rotation, translation=[0, 0, 0], verbose=False
):
    """Predict on rigidly transformed features.

    The rigid transformation is applied to the volumes prior to prediction, and
    the prediced labels are transformed with the inverse warp, so that they are
    in the same space.

    Parameters
    ----------
    model: `tf.keras.Model`, model used for prediction.
    x: 3D array, volume of features.
    block_shape: tuple of length 3, shape of non-overlapping blocks to take
        from the features. This also corresponds to the input of the model, not
        including the batch or channel dimensions.
    rotation: tuple of length 3, rotation angle in radians in each dimension.
    translation: tuple of length 3, units of translation in each dimension.
    verbose: bool, whether to print progress bar.

    Returns
    -------
    Array of predictions with the same shape and in the same space as the
    original input features.
    """

    x = np.asarray(x).astype(np.float32)
    affine = get_affine(x.shape, rotation=rotation, translation=translation)
    inverse_affine = tf.linalg.inv(affine)
    x_warped = warp(x, affine, order=1)

    x_warped_blocks = to_blocks_numpy(x_warped, block_shape)
    x_warped_blocks = x_warped_blocks[..., np.newaxis]  # add grayscale channel
    x_warped_blocks = standardize_numpy(x_warped_blocks)
    y = model.predict(x_warped_blocks, batch_size=1, verbose=verbose)

    n_classes = y.shape[-1]
    if n_classes == 1:
        y = y.squeeze(-1)
    else:
        # Usually, the argmax would be taken to get the class membership of
        # each voxel, but if we get hard values, then we cannot average
        # multiple predictions.
        raise ValueError(
            "This function is not compatible with multi-class predictions."
        )

    y = from_blocks_numpy(y, x.shape)
    y = warp(y, inverse_affine, order=0).numpy()

    return y


def predict_by_estimator(
    filepath,
    model_path,
    block_shape,
    batch_size=4,
    normalizer=None,
    n_samples=1,
    return_variance=True,
    return_entropy=True,
):
    """Predict on a volume given a filepath and a path to tensorflow1 estimator
    saved model.

    Parameters
    ----------
    filepath: path-like, path to existing neuroimaging volume on which
        to predict.
    model_path: path_like, path to saved tf1 estimator model, trained model.
    block_shape: tuple of length 3, shape of sub-volumes on which to
        predict.
    batch_size: int, number of sub-volumes per batch for predictions.
    normalizer: callable, function that accepts an ndarray and returns an
        ndarray. Called before separating volume into blocks.
    n_samples: The number of sampling. If set as 1, it will just return the
        single prediction value. The default value is 1
    return_variance: Boolean. If set True, it returns the running population
        variance along with mean. Note, if the n_samples is smaller or equal to 1,
        the variance will not be returned; instead it will return None
    return_entropy: Boolean. If set True, it returns the running entropy.
        along with mean.

    Returns
    -------
    `nibabel.spatialimages.SpatialImage` or arrays of predictions of
        mean, variance(optional), and entropy (optional).
    """

    if not Path(filepath).is_file():
        raise FileNotFoundError("could not find file {}".format(filepath))

    img = nib.load(filepath)
    inputs = np.asarray(img.dataobj)
    img.uncache()
    inputs = inputs.astype(np.float32)

    if normalizer:
        features = normalizer(inputs)
    else:
        features = inputs

    features = to_blocks_numpy(features, block_shape=block_shape)

    # Add a dimension for single channel.
    features = features[..., None]

    # restores the tf estimator model
    predictor = _get_predictor(model_path)

    # Predict per block to reduce memory consumption.
    n_blocks = features.shape[0]
    n_batches = math.ceil(n_blocks / batch_size)

    # Variational inference
    means = np.zeros_like(features.squeeze(-1))
    variances = np.zeros_like(features.squeeze(-1))
    entropies = np.zeros_like(features.squeeze(-1))
    progbar = tf.keras.utils.Progbar(n_batches)
    progbar.update(0)
    for j in range(0, n_blocks, batch_size):
        this_x = tf.convert_to_tensor(features[j : j + batch_size])
        s = StreamingStats()
        for n in range(n_samples):
            new_prediction = predictor(this_x)
            s.update(new_prediction["probabilities"])

        means[j : j + batch_size] = np.argmax(s.mean(), axis=-1)  # max mean
        variances[j : j + batch_size] = np.sum(s.var(), axis=-1)
        entropies[j : j + batch_size] = np.sum(s.entropy(), axis=-1)  # entropy
        progbar.add(1)

        total_means = from_blocks(means, output_shape=inputs.shape).numpy()
        total_variance = from_blocks(variances, output_shape=inputs.shape).numpy()
        total_entropy = from_blocks(entropies, output_shape=inputs.shape).numpy()

    include_variance = (n_samples > 1) and (return_variance)
    if include_variance:
        if return_entropy:
            # return in the form of a nifti image
            total_means = nib.spatialimages.SpatialImage(
                dataobj=total_means,
                affine=img.affine,
                header=img.header,
                extra=img.extra,
            )

            total_variance = nib.spatialimages.SpatialImage(
                dataobj=total_variance,
                affine=img.affine,
                header=img.header,
                extra=img.extra,
            )

            total_entropy = nib.spatialimages.SpatialImage(
                dataobj=total_entropy,
                affine=img.affine,
                header=img.header,
                extra=img.extra,
            )
            return total_means, total_variance, total_entropy
        else:
            total_means = nib.spatialimages.SpatialImage(
                dataobj=total_means,
                affine=img.affine,
                header=img.header,
                extra=img.extra,
            )

            total_variance = nib.spatialimages.SpatialImage(
                dataobj=total_variance,
                affine=img.affine,
                header=img.header,
                extra=img.extra,
            )
            return total_means, total_variance
    else:
        if return_entropy:
            total_means = nib.spatialimages.SpatialImage(
                dataobj=total_means,
                affine=img.affine,
                header=img.header,
                extra=img.extra,
            )

            total_entropy = nib.spatialimages.SpatialImage(
                dataobj=total_entropy,
                affine=img.affine,
                header=img.header,
                extra=img.extra,
            )
            return total_means, total_entropy
        else:
            total_means = nib.spatialimages.SpatialImage(
                dataobj=total_means,
                affine=img.affine,
                header=img.header,
                extra=img.extra,
            )
            return total_means



"""Setup script for nobrainer.

To install, run `python3 setup.py install`.
"""
import os
import sys

from setuptools import setup

# This is needed for versioneer to be importable when building with PEP 517.
# See <https://github.com/warner/python-versioneer/issues/193> and links
# therein for more information.
sys.path.append(os.path.dirname(__file__))

try:
    import versioneer

    setup_kw = {
        "version": versioneer.get_version(),
        "cmdclass": versioneer.get_cmdclass(),
    }
except ImportError:
    # see https://github.com/warner/python-versioneer/issues/192
    print("WARNING: failed to import versioneer, falling back to no version for now")
    setup_kw = {}

setup(name="nobrainer", **setup_kw)



"""Input/output methods."""
import csv
import functools
import multiprocessing
import os

from fsspec.implementations.local import LocalFileSystem
import nibabel as nib
import numpy as np
import tensorflow as tf

from .utils import get_num_parallel

_TFRECORDS_FEATURES_DTYPE = "float32"


def read_csv(filepath, skip_header=True, delimiter=","):
    """Return list of tuples from a CSV, where each tuple contains the items
    in a row.
    """
    with open(filepath, newline="") as csvfile:
        reader = csv.reader(csvfile, delimiter=delimiter)
        if skip_header:
            next(reader)
        return [tuple(row) for row in reader]


def read_mapping(filepath, skip_header=True, delimiter=","):
    """Read CSV to dictionary, where first column becomes keys and second
    columns becomes values. Other columns are ignored. Keys and values are
    coerced to integers.
    """
    mapping = read_csv(filepath, skip_header=skip_header, delimiter=delimiter)
    if not all(map(lambda r: len(r) >= 2, mapping)):
        raise ValueError("not all rows in the mapping have at least 2 values")
    try:
        return {int(row[0]): int(row[1]) for row in mapping}
    except ValueError:
        raise ValueError("mapping values must be integers but non-integer encountered")


def read_volume(filepath, dtype=None, return_affine=False, to_ras=False):
    """Return numpy array of data from a neuroimaging file."""
    img = nib.load(filepath)
    if to_ras:
        img = nib.as_closest_canonical(img)
    data = img.get_fdata(caching="unchanged")
    if dtype is not None:
        data = data.astype(dtype)
    return data if not return_affine else (data, img.affine)


def verify_features_labels(
    volume_filepaths,
    volume_shape=(256, 256, 256),
    check_shape=True,
    check_labels_int=True,
    check_labels_gte_zero=True,
    num_parallel_calls=None,
    verbose=1,
):
    """Verify a list of files. This function is meant to be run before
    converting volumes to TFRecords.

    Parameters
    ----------
    volume_filepaths: nested list. Every sublist in the list should contain two
        items: (1) path to feature volume and (2) path to label volume or a scalar.
    volume_shape: tuple of three ints. Shape that both volumes should be.
    check_shape: boolean, if true, validate that the shape of both volumes is
        equal to 'volume_shape'.
    check_labels_int: boolean, if true, validate that every labels volume is an
        integer type or can be safely converted to an integer type.
    check_labels_gte_zero: boolean, if true, validate that every labels volume
        has values greater than or equal to zero.
    num_parallel_calls: int, number of processes to use for multiprocessing. If
        None, will use all available processes.
    verbose: {0, 1, 2}, verbosity of the progress bar. 0 is silent, 1 is verbose,
        and 2 is semi-verbose.

    Returns
    -------
    List of invalid pairs of filepaths. If the list is empty, all filepaths are
    valid.
    """
    from nobrainer.tfrecord import _labels_all_scalar

    for pair in volume_filepaths:
        if len(pair) != 2:
            raise ValueError(
                "all items in 'volume_filepaths' must have length of 2, but"
                " found at least one item with length != 2."
            )

    labels = (y for _, y in volume_filepaths)
    scalar_labels = _labels_all_scalar(labels)

    for pair in volume_filepaths:
        if not os.path.exists(pair[0]):
            raise ValueError("file does not exist: {}".format(pair[0]))
        if not scalar_labels:
            if not os.path.exists(pair[1]):
                raise ValueError("file does not exist: {}".format(pair[1]))

    if scalar_labels:
        map_fn = functools.partial(
            _verify_features_scalar_labels,
            volume_shape=volume_shape,
            check_shape=check_shape,
        )
    else:
        map_fn = functools.partial(
            _verify_features_nonscalar_labels,
            volume_shape=volume_shape,
            check_shape=check_shape,
            check_labels_int=check_labels_int,
            check_labels_gte_zero=check_labels_gte_zero,
        )
    if num_parallel_calls is None:
        num_parallel_calls = get_num_parallel()

    print("Verifying {} examples".format(len(volume_filepaths)))
    progbar = tf.keras.utils.Progbar(len(volume_filepaths), verbose=verbose)
    progbar.update(0)

    outputs = []
    if num_parallel_calls == 1:
        for vf in volume_filepaths:
            valid = map_fn(vf)
            outputs.append(valid)
            progbar.add(1)
    else:
        with multiprocessing.Pool(num_parallel_calls) as p:
            for valid in p.imap(map_fn, volume_filepaths, chunksize=2):
                outputs.append(valid)
                progbar.add(1)
    invalid_files = [
        pair for valid, pair in zip(outputs, volume_filepaths) if not valid
    ]
    return invalid_files


def _verify_features_nonscalar_labels(
    pair_of_paths, *, volume_shape, check_shape, check_labels_int, check_labels_gte_zero
):
    """Verify a pair of features and labels volumes."""
    x = nib.load(pair_of_paths[0])
    y = nib.load(pair_of_paths[1])
    if check_shape:
        if not volume_shape:
            raise ValueError(
                "`volume_shape` must be specified if `check_shape` is true."
            )
        if x.shape != volume_shape:
            return False
        if x.shape != y.shape:
            return False
    if check_labels_int:
        # Quick check of integer type.
        if not np.issubdtype(y.dataobj.dtype, np.integer):
            return False
        y = y.get_fdata(caching="unchanged", dtype=np.float32)
        # Longer check that all values in labels can be cast to int.
        if not np.all(np.mod(y, 1) == 0):
            return False
    if check_labels_gte_zero:
        if not np.all(y >= 0):
            return False
    return True


def _verify_features_scalar_labels(path_scalar, *, volume_shape, check_shape):
    """Check that feature has the desired shape and that label is scalar."""
    from nobrainer.tfrecord import _is_int_or_float

    feature, label = path_scalar
    x = nib.load(feature)
    if check_shape:
        if not volume_shape:
            raise ValueError(
                "`volume_shape` must be specified if `check_shape` is true."
            )
        if x.shape != volume_shape:
            return False
    if not _is_int_or_float(label):
        return False
    return True


def _is_gzipped(filepath, filesys=None):
    """Return True if the file is gzip-compressed, False otherwise."""
    fs = filesys if filesys is not None else LocalFileSystem()
    with fs.open(filepath, "rb") as f:
        return f.read(2) == b"\x1f\x8b"



import tensorflow as tf


def centercrop(x, y=None, finesize=64, trans_xy=False):
    """Apply center crop to input and label.

    Usage:
    ```python
    >>> x = np.array([[[1, 1, 1], [2, 2, 2], [3, 3, 3]]])
    >>> finesize = 1
    >>> x_out = spatial_transforms.centercrop(x, finesize=finesize)
    >>> x_out
        <tf.Tensor: shape=(1, 1, 3), dtype=float32,
        numpy=array([[[2., 2., 2.]]], dtype=float32)>
    ```

    Parameters
    ----------
        x: input is a tensor or numpy to have rank 3,
        y: label is a tensor or numpy to have rank 3,
        finesize: int, desired output size of the crop. Default = 64;
        trans_xy: Boolean, transforms both x and y (Default: False).
           If set True, function will require both x,y.

    Returns
    ----------
        CenterCroped input and/or label tensor.
    """
    if ~tf.is_tensor(x):
        x = tf.convert_to_tensor(x)
    x = tf.cast(x, tf.float32)
    if len(x.shape) != 3:
        raise ValueError("`volume` must be rank 3")
    w, h = x.shape[1], x.shape[0]
    th, tw = finesize, finesize
    x1 = int(round((w - tw) / 2.0))
    y1 = int(round((h - th) / 2.0))
    x = x[y1 : y1 + th, x1 : x1 + tw, :]

    if trans_xy:
        if y is None:
            raise ValueError("`LabelMap' should be assigned")
        if len(y.shape) != 3:
            raise ValueError("`LabelMap` must be equal or higher than rank 2")
        if ~tf.is_tensor(y):
            y = tf.convert_to_tensor(y)
        y = tf.cast(y, tf.float32)
        y = y[y1 : y1 + th, x1 : x1 + tw, :]
        return x, y
    else:
        return x


def spatialConstantPadding(x, y=None, trans_xy=False, padding_zyx=[1, 1, 1]):
    """Add constant padding to input and label.

    Usage:
    ```python
    >>> x = np.array([[[1, 1, 1], [2, 2, 2], [3, 3, 3]]])
    >>> x_out = spatial_transforms.spatialConstantPadding(
    x,padding_zyx=[0, 1, 1])
    >>> x_out
    <tf.Tensor: shape=(1, 5, 5), dtype=float32, numpy=
    array([[[0., 0., 0., 0., 0.],
            [0., 1., 1., 1., 0.],
            [0., 2., 2., 2., 0.],
            [0., 3., 3., 3., 0.],
            [0., 0., 0., 0., 0.]]], dtype=float32)>
    ```

    Parameters
    ----------
        x: input is a tensor or numpy to have rank 3,
        y: label is a tensor or numpy to have rank 3,
        padding_zyx: int or a list of desired padding in three dimensions.
            Default = 1;
        trans_xy: Boolean, transforms both x and y (Default: False).
           If set True, function will require both x,y.

    Returns
    ----------
        Input and/or label tensor with spatial padding.
    """
    if ~tf.is_tensor(x):
        x = tf.convert_to_tensor(x)
    x = tf.cast(x, tf.float32)
    padz = padding_zyx[0]
    pady = padding_zyx[1]
    padx = padding_zyx[2]
    padding = tf.constant([[padz, padz], [pady, pady], [padx, padx]])
    x = tf.pad(x, padding, "CONSTANT")
    if trans_xy:
        if y is None:
            raise ValueError("`LabelMap' should be assigned")
        if len(y.shape) != 3:
            raise ValueError("`LabelMap` must be equal or higher than rank 2")
        if ~tf.is_tensor(y):
            y = tf.convert_to_tensor(y)
        y = tf.cast(y, tf.float32)
        y = tf.pad(y, padding, "CONSTANT")
        return x, y
    else:
        return x


def randomCrop(x, y=None, trans_xy=False, cropsize=16):
    """Apply random crops to input and label.

    Usage:
    ```python
    >>> x = np.array([[[1, 1, 1], [2, 2, 2], [3, 3, 3]]])
    >>> x_out = spatial_transforms.randomCrop(x, cropsize=1)
    >>> x_out
    <tf.Tensor: shape=(1, 1, 3), dtype=float32,
    numpy=array([[[6., 2., 5.]]], dtype=float32)>
    ```

    Parameters
    ----------
        x: input is a tensor or numpy to have rank 3,
        y: label is a tensor or numpy to have rank 3,
        cropsize: int, the size of the cropped output,
        finesize: int, desired output size of the crop. Default = 64;
        trans_xy: Boolean, transforms both x and y (Default: False).
           If set True, function will require both x,y.

    Returns
    ----------
        Randomly croped input and/or label tensor.
    """
    if ~tf.is_tensor(x):
        x = tf.convert_to_tensor(x)
    x = tf.cast(x, tf.float32)
    if trans_xy:
        if y is None:
            raise ValueError("`LabelMap' should be assigned")
        if len(y.shape) != 3:
            raise ValueError("`LabelMap` must be equal or higher than rank 2")
        if ~tf.is_tensor(y):
            y = tf.convert_to_tensor(y)
        y = tf.cast(y, tf.float32)
        stacked = tf.stack([x, y], axis=0)
        cropped = tf.image.random_crop(stacked, [2, cropsize, cropsize, x.shape[2]])
        return cropped[0], cropped[1]
    else:
        return tf.image.random_crop(x, [cropsize, cropsize, x.shape[2]])


def resize(x, y=None, trans_xy=False, size=[32, 32], mode="bicubic"):
    """Resize the input and label.

    Usage:
    ```python
     >>> x = np.array([[[1, 2, 3], [6, 2, 5], [3, 4, 9]]])
     >>> x_out = spatial_transforms.resize(x,size=[2, 2])
     >>> x_out
     <tf.Tensor: shape=(2, 2, 3), dtype=float32, numpy=
     array([[[2.0145986, 1.9562043, 3.2919707],
             [3.678832 , 3.620438 , 8.284672 ]],

            [[2.0145986, 1.9562043, 3.2919707],
             [3.678832 , 3.620438 , 8.284672 ]]], dtype=float32)>
     ```

     Parameters
     ----------
        x: input is a tensor or numpy to have rank 3,
        y: label is a tensor or numpy to have rank 3,
        size: int or a list, the resize output,
        trans_xy: Boolean, transforms both x and y (Default: False).
           If set True, function will require both x,y.
        mode [options]: "bilinear", "lanczos3",
        "lanczos5", "bicubic", "gaussian" , "nearest".

    Returns
    ----------
         Resized input and/or label tensor.
    """
    if ~tf.is_tensor(x):
        x = tf.convert_to_tensor(x)
    x = tf.cast(x, tf.float32)
    x = tf.image.resize(x, size, method=mode)
    if trans_xy:
        if y is None:
            raise ValueError("`LabelMap' should be assigned")
        if len(y.shape) != 3:
            raise ValueError("`LabelMap` must be equal or higher than rank 2")
        if ~tf.is_tensor(y):
            y = tf.convert_to_tensor(y)
        y = tf.cast(y, tf.float32)
        y = tf.image.resize(y, size, method=mode)
        return x, y
    else:
        return x


def randomflip_leftright(x, y=None, trans_xy=False):
    """Randomly flips the input and label.

    Usage:
    ```python
    >>> x = np.array([[[1, 2, 3], [6, 2, 5], [3, 4, 9]]])
    >>> x_out = spatial_transforms.randomflip_leftright(x)
    >>> x_out
    <tf.Tensor: shape=(1, 3, 3), dtype=float32, numpy=
    array([[[3., 4., 9.],
            [6., 2., 5.],
            [1., 2., 3.]]], dtype=float32)>
    ```

    Parameters
    ----------
        x: input is a tensor or numpy to have rank 3,
        y: label is a tensor or numpy to have rank 3,
        trans_xy: Boolean, transforms both x and y (Default: False).
           If set True, function will require both x,y.

    Returns
    ----------
        Randomly flipped input and/or label tensor.
    """
    if ~tf.is_tensor(x):
        x = tf.convert_to_tensor(x)
    x = tf.cast(x, tf.float32)
    if trans_xy:
        if y is None:
            raise ValueError("`LabelMap' should be assigned")
        if len(y.shape) != 3:
            raise ValueError("`LabelMap` must be equal or higher than rank 2")
        if ~tf.is_tensor(y):
            y = tf.convert_to_tensor(y)
        y = tf.cast(y, tf.float32)
        c = tf.concat([x, y], axis=0)
        c = tf.image.random_flip_left_right(c, seed=None)
        split_channel = int(c.shape[0] / 2)
        return c[0:split_channel, :, :], c[split_channel : c.shape[0], :, :]
    else:
        return tf.image.random_flip_left_right(x, seed=None)
